# Copyright 2021 The KCL Authors. All rights reserved.

from typing import Optional, List, Tuple
from dataclasses import dataclass, field

from .code import Opcode


@dataclass
class OpcodeContent:
    op: Opcode
    arg: Optional[int]
    arg_list: List[int]
    index: int
    meta: Tuple[str, int, int]

    @property
    def begin_index(self):
        return self.index

    @property
    def end_index(self):
        return self.index + len(self.arg_list) + 1


@dataclass
class OpcodeFactory:
    pkgpath: Optional[str]
    values: List[OpcodeContent] = field(default_factory=list)

    @property
    def begin_index(self):
        return self.values[0].begin_index

    @property
    def end_index(self):
        return self.values[-1].end_index

    def pretty_print(self):
        if not self.values:
            return
        for value in self.values:
            print(
                ">>> ",
                value.op,
                "arg:",
                value.arg,
                "index:",
                value.index,
                "line",
                value.meta[1],
            )

    # ---------------------
    # Static methods
    # ---------------------

    @staticmethod
    def build_from_codes(codes: list, pkgpath: str) -> "OpcodeFactory":
        f = OpcodeFactory(pkgpath=pkgpath)
        isp = 0
        while codes and isp < len(codes):
            code = codes[isp]
            opcode_index = isp
            isp += 1
            # Skip the invalid opcode
            if code == Opcode.INVALID:
                continue
            arg = None
            arg_list = []
            if Opcode.has_arg(code):
                arg = codes[isp] + (codes[isp + 1] << 8) + (codes[isp + 2] << 16)
                arg_list = [codes[isp], codes[isp + 1], codes[isp + 2]]
                isp += 3
            info = codes[isp]
            isp += 1
            f.values.append(
                OpcodeContent(
                    op=Opcode(code),
                    arg=arg,
                    arg_list=arg_list,
                    index=opcode_index,
                    meta=info,
                )
            )
        return f

    @staticmethod
    def to_codes(contents: List[OpcodeContent]) -> list:
        return sum(
            [
                [content.op]
                + (
                    ([content.op] * len(content.arg_list or []))
                    if content.op == Opcode.INVALID
                    else (content.arg_list or [])
                )
                + [content.op if content.op == Opcode.INVALID else content.meta]
                for content in contents or []
            ],
            [],
        )


@dataclass
class SchemaBodyOpcodeFactory(OpcodeFactory):
    schema_name: str = None
    begin_code: List[OpcodeContent] = None
    end_code: List[OpcodeContent] = None

    def validate(self) -> bool:
        return (
            self.values
            and self.values[0].op == Opcode.STORE_LOCAL
            and self.values[-1].op == Opcode.RETURN_VALUE
        )

    def update_start_end_code(self) -> "SchemaBodyOpcodeFactory":
        if not self.validate():
            raise ValueError("Invalid schema opcode factory")
        for i, value in enumerate(self.values):
            if value.op == Opcode.SCHEMA_NOP and not self.begin_code:
                self.begin_code = self.values[:i]
            if value.op == Opcode.RETURN_VALUE:
                self.end_code = [value]
        return self

    def split_to_schema_attr_codes(self) -> List["SchemaBodyOpcodeFactory"]:
        if not self.validate():
            raise ValueError("Invalid schema opcode factory")
        results = []
        last_nop_index = 0
        next_nop_index = -1
        if_stack = [False]
        if not self.end_code or not self.begin_code:
            self.update_start_end_code()
        for i, value in enumerate(self.values):
            if value.op == Opcode.POP_JUMP_IF_FALSE:
                if_stack.append(True)
            if value.op == Opcode.JUMP_FORWARD:
                if_stack.pop()
            if value.op == Opcode.SCHEMA_NOP:
                last_nop_index = i
            if SchemaBodyOpcodeFactory.is_schema_line_op(value.op):
                for j, v in enumerate(self.values[i + 1 :]):
                    if v.op == Opcode.SCHEMA_NOP:
                        next_nop_index = i + j + 1
                        break
                values_append = []
                if if_stack[-1]:
                    for v in self.values[last_nop_index : next_nop_index + 1]:
                        if value.index < v.index < self.values[next_nop_index].index:
                            values_append.append(
                                OpcodeContent(
                                    op=Opcode.INVALID,
                                    arg=v.arg,
                                    arg_list=v.arg_list,
                                    index=v.index,
                                    meta=v.meta,
                                )
                            )
                        else:
                            values_append.append(v)
                else:
                    values_append = self.values[last_nop_index : next_nop_index + 1]
                results.append(
                    SchemaBodyOpcodeFactory(
                        pkgpath=self.pkgpath,
                        schema_name=self.schema_name,
                        values=values_append,
                        begin_code=self.begin_code,
                        end_code=self.end_code,
                    )
                )
        return results

    def to_run_code_list(self) -> list:
        assert self.end_code and self.begin_code
        # Single attribute code
        begin_codes = OpcodeFactory.to_codes(self.begin_code)
        end_codes = OpcodeFactory.to_codes(self.end_code)
        codes = OpcodeFactory.to_codes(self.values)
        start_to_code_count = self.values[0].index - self.begin_code[-1].end_index - 1
        code_to_end_count = self.end_code[0].index - self.values[-1].end_index - 1
        return (
            begin_codes
            + [Opcode.INVALID] * start_to_code_count
            + codes
            + [Opcode.INVALID] * code_to_end_count
            + end_codes
        )

    def pretty_print(self):
        if not self.values:
            return
        RED = "\033[31m"
        GREEN = "\033[32m"
        BLUE = "\033[34m"
        BOLD = "\033[1m"
        RESET = "\033[m"
        found_nop = False
        for value in self.values:
            if value.op == Opcode.SCHEMA_NOP:
                found_nop = True
                print(f"{BOLD}{GREEN}", end="")
            elif value.op == Opcode.RETURN_VALUE or not found_nop:
                print(f"{BOLD}{RED}", end="")
            else:
                print(f"{BOLD}{BLUE}", end="")
            print(
                ">>> ",
                value.op,
                "arg:",
                value.arg,
                "index:",
                value.index,
                "line",
                value.meta[1],
            )
        print(f"{RESET}")

    # ---------------------
    # Static methods
    # ---------------------

    @staticmethod
    def is_schema_line_op(op: int):
        return (
            op == Opcode.SCHEMA_ATTR
            or op == Opcode.SCHEMA_UPDATE_ATTR
            or op == Opcode.STORE_ATTR
        )

    @staticmethod
    def build_from_codes(
        codes: list, pkgpath: str, schema_name: str
    ) -> "SchemaBodyOpcodeFactory":
        ctx = OpcodeFactory.build_from_codes(codes, pkgpath)
        return SchemaBodyOpcodeFactory(
            pkgpath=ctx.pkgpath,
            schema_name=schema_name,
            values=ctx.values,
        )



import io

import kclvm.compiler.extension.builtin.system_module.yaml as yaml


def KMANGLED_yaml_stream(data: list, opts: dict = {}) -> None:
    """This function is used to serialize the KCL object list into YAML output with the --- separator. It has two parameters:


    Parameters
    ----------
    + values - A list of KCL objects
    + opts - The YAML serialization options
        + sort_keys: Whether to sort the serialized data in the dictionary order of attribute names (the default is False).
        + ignore_private: Whether to ignore the attribute output whose name starts with the character _ (the default value is True).
        + ignore_none: Whether to ignore the attribute with the value of' None '(the default value is False).
        + sep: Set the separator between multiple YAML documents (the default value is "---").
    """
    if not data:
        return None
    sort_keys = opts.get("sort_keys", False)
    ignore_private = opts.get("ignore_private", True)
    ignore_none = opts.get("ignore_none", False)
    sep = opts.get("sep", "---")
    outputs = ""
    with io.StringIO() as buf:
        if data:
            for result in data[:-1]:
                if result:
                    buf.write(
                        yaml.KMANGLED_encode(
                            result,
                            sort_keys=sort_keys,
                            ignore_private=ignore_private,
                            ignore_none=ignore_none,
                        )
                    )
                    buf.write(f"{sep}\n")
            if data[-1]:
                buf.write(
                    yaml.KMANGLED_encode(
                        data[-1],
                        sort_keys=sort_keys,
                        ignore_private=ignore_private,
                        ignore_none=ignore_none,
                    )
                )
            outputs = buf.getvalue()

    from kclvm.vm.vm import VirtualMachine

    VirtualMachine.hook_output = list(yaml.KMANGLED_load_all(outputs))
    return None



# Copyright 2021 The KCL Authors. All rights reserved.

from io import StringIO
from dataclasses import dataclass, field
from typing import cast, List, Optional, Union

import kclvm.kcl.ast as ast
import kclvm.compiler.astutil as astutil


VERTEX_ROOT_NAME = "@root"


@dataclass
class Meta:
    """The filename, line and column info"""

    filename: Optional[str] = field(default_factory=lambda: None)
    line: Optional[int] = field(default_factory=lambda: None)
    column: Optional[int] = field(default_factory=lambda: None)
    end_line: Optional[int] = field(default_factory=lambda: None)
    end_column: Optional[int] = field(default_factory=lambda: None)

    @staticmethod
    def from_ast_node(node: ast.AST) -> "Meta":
        return (
            Meta(
                filename=node.filename,
                line=node.line,
                column=node.column,
                end_line=node.end_line,
                end_column=node.end_column,
            )
            if node and isinstance(node, ast.AST)
            else Meta()
        )


@dataclass
class ConfigMeta:
    """The schema config name, args and kwargs AST node references"""

    name: str = None
    pkgpath: str = None
    args: List[ast.Expr] = None
    kwargs: List[ast.Keyword] = None


@dataclass
class VertexOption:
    """The vertex unification options"""

    is_union: bool = False  # `:`
    is_override: bool = False  # `=`
    is_append: bool = False  # `+=`
    is_unique: bool = False  # `!`
    is_strategy_merge: bool = False  # `?`


@dataclass
class Vertex:
    """
    Parameters
    ----------
    name: Union[int, float, str, ast.AST]
        Vertex node name. When the name is a variable needed to be calculated in VM,
        it is a AST type
    adjs: List["Vertex"]
        The list of downstream nodes of the vertex
    node: ast.AST
        The AST node reference
    meta: Mate
        The filename, line and column info
    config_meta: ConfigMeta:
        The schema config name, args and kwargs AST node references
    option:
        The vertex unification options

    Methods
    -------
    vertex_to_ast:
        Vertex -> AST
    ast_to_vertex:
        AST -> Vertex
    """

    name: Union[int, float, str, ast.AST]
    adjs: Optional[List["Vertex"]]
    node: ast.AST = field(default_factory=lambda: None)
    meta: Optional[Meta] = field(default_factory=lambda: Meta)
    config_meta: Optional[ConfigMeta] = field(default_factory=lambda: ConfigMeta)
    option: Optional[VertexOption] = field(default_factory=lambda: VertexOption)

    # ---------------
    # Member method
    # ---------------

    def pretty(self, indent: int = 1) -> str:
        """Pretty print to show the vertex structure"""
        with StringIO() as buf:
            buf.write("name: " + str(self.name) + "\n")
            buf.write("is_unique: " + str(self.option.is_unique) + "\n")
            if self.config_meta.name:
                buf.write("config_name: " + self.config_meta.name + "\n")
            if self.adjs:
                buf.write("adjs: \n")
                for v in self.adjs:
                    lines = v.pretty(indent).split("\n")
                    buf.write(
                        "\n".join([" " * indent * 4 + line for line in lines]) + "\n"
                    )
            else:
                buf.write("value: " + str(self.node))
            return buf.getvalue().rstrip(" \n") + "\n"

    def vertex_to_ast(self) -> Optional[ast.AST]:
        """Vertex to KCL AST"""

        def append_config_key_value(
            t: Union[ast.SchemaExpr, ast.ConfigExpr],
            key: Union[int, float, str, Optional[ast.AST]],
            value: ast.AST,
            op: int = ast.ConfigEntryOperation.UNION,
            meta: Meta = Meta(),
        ):
            if isinstance(t, ast.SchemaExpr):
                if not t.config:
                    t.config = ast.ConfigExpr(line=t.line, column=t.column)
                    t.config.filename = meta.filename
                if not t.config.items:
                    t.config.items = []
                # If `key` is None, it may be a double star expr
                key_node = None
                if key and isinstance(key, ast.AST):
                    key_node = key
                elif isinstance(key, str):
                    key_node = ast.Identifier(
                        line=meta.line, column=meta.column, names=[key]
                    ).set_filename(meta.filename)
                    key_node.end_line, key_node.end_column = (
                        meta.end_line,
                        meta.end_column,
                    )
                elif isinstance(key, (int, float)):
                    key_node = ast.NumberLit(
                        line=meta.line, column=meta.column, value=key
                    )
                    key_node.end_line, key_node.end_column = (
                        meta.end_line,
                        meta.end_column,
                    )
                if isinstance(key_node, ast.AST):
                    key_node.filename = meta.filename
                value = cast(ast.Expr, value)
                t.config.items.append(
                    ast.ConfigEntry(
                        key=key_node,
                        value=value,
                        operation=op,
                    )
                )
            elif isinstance(t, ast.ConfigExpr):
                if not t.items:
                    t.items = []
                # If `key` is None, it may be a double star expr
                key_node = None
                if key and isinstance(key, ast.AST):
                    key_node = key
                elif isinstance(key, str):
                    key_node = ast.StringLit(
                        line=meta.line, column=meta.column, value=key
                    )
                    key_node.end_line, key_node.end_column = (
                        meta.end_line,
                        meta.end_column,
                    )
                elif isinstance(key, (int, float)):
                    key_node = ast.NumberLit(
                        line=meta.line, column=meta.column, value=key
                    )
                    key_node.end_line, key_node.end_column = (
                        meta.end_line,
                        meta.end_column,
                    )
                if isinstance(key_node, ast.AST):
                    key_node.filename = meta.filename
                key_node = cast(ast.Expr, key_node)
                t.items.append(
                    ast.ConfigEntry(
                        key=key_node,
                        value=value,
                        operation=op,
                    )
                )

        # Get root vertex
        if isinstance(self.name, str) and self.name == VERTEX_ROOT_NAME:
            module = ast.Module(filename=self.meta.filename, line=1, column=1)
            for v in self.adjs:
                assign_stmt = ast.AssignStmt(
                    line=v.meta.line,
                    column=v.meta.column,
                )
                assign_stmt.targets = [
                    ast.Identifier(
                        line=v.node.line,
                        column=v.node.column,
                        names=[v.name],
                        ctx=ast.ExprContext.STORE,
                    ).set_filename(v.node.filename)
                ]
                assign_stmt.end_line = v.meta.end_line
                assign_stmt.end_column = v.meta.end_column
                assign_stmt.value = v.vertex_to_ast()
                assign_stmt.filename = v.meta.filename
                module.body.append(assign_stmt)
            return module
        # Get normal vertex such as in the right assignment
        else:
            # SchemaExpr config
            if self.config_meta.name:
                config_expr = ast.SchemaExpr(
                    line=self.meta.line,
                    column=self.meta.column,
                )
                name = ast.Identifier(
                    line=self.meta.line,
                    column=self.meta.column,
                    names=self.config_meta.name.split("."),
                ).set_filename(self.meta.filename)
                name.end_line, name.end_column = (
                    self.meta.end_line,
                    self.meta.end_column,
                )
                config_expr.name = name
                config_expr.name.pkgpath = self.config_meta.pkgpath
                config_expr.args = self.config_meta.args
                config_expr.kwargs = self.config_meta.kwargs
            # ConfigExpr config
            else:
                config_expr = ast.ConfigExpr(
                    line=self.meta.line,
                    column=self.meta.column,
                )
            config_expr.end_line, config_expr.end_column = (
                self.meta.end_line,
                self.meta.end_column,
            )
            if self.adjs:
                for vv in self.adjs:
                    op = ast.ConfigEntryOperation.UNION
                    op = (
                        ast.ConfigEntryOperation.OVERRIDE
                        if vv.option.is_override
                        else op
                    )
                    op = ast.ConfigEntryOperation.INSERT if vv.option.is_append else op
                    if vv.adjs:
                        append_config_key_value(
                            config_expr, vv.name, vv.vertex_to_ast(), op, vv.meta
                        )
                    elif vv.node:
                        append_config_key_value(
                            config_expr, vv.name, vv.node, op, vv.meta
                        )
                return config_expr
            elif self.node:
                return self.node
            return None

    # ---------------
    # Static method
    # ---------------

    @staticmethod
    def update_vertex_option(v: "Vertex", op: int):
        """Update the vertex option using the schema config operation"""
        if not isinstance(v, Vertex):
            return
        v.option = VertexOption(
            is_override=op == ast.ConfigEntryOperation.OVERRIDE,
            is_append=op == ast.ConfigEntryOperation.INSERT,
            is_union=op == ast.ConfigEntryOperation.UNION,
        )

    @staticmethod
    def ast_to_vertex(
        t: ast.AST,
        name: Optional[Union[int, float, str, ast.AST]] = None,
        is_in_schema: bool = False,
    ) -> Optional["Vertex"]:
        """Build a vertex from AST"""
        if not t or not isinstance(t, ast.AST):
            return None
        if isinstance(t, ast.Module):
            t = cast(ast.Module, t)
            root = Vertex.new_root(
                node=t,
                adjs=[],
            )
            declarations = astutil.filter_declarations(t)
            for d in declarations:
                if "." not in d.name:
                    vertex = Vertex.ast_to_vertex(d.value, d.name)
                    vertex.meta.filename = d.filename
                    vertex.option.is_union = d.is_union
                    root.adjs.append(vertex)
            return root
        elif isinstance(t, ast.SchemaExpr):
            vertex = Vertex.ast_to_vertex(t.config, name, True)
            vertex.meta = Meta.from_ast_node(t)
            vertex.name = name
            vertex.config_meta = ConfigMeta(
                name=t.name.get_name(),
                pkgpath=t.name.pkgpath,
                args=t.args,
                kwargs=t.kwargs,
            )
            if not vertex.adjs and isinstance(vertex.node, ast.ConfigExpr):
                vertex.node = t
            return vertex
        elif isinstance(t, ast.ConfigExpr):
            vertex = Vertex(
                node=t,
                name=name,
                adjs=[],
                meta=Meta.from_ast_node(t),
            )
            for key, value, operation in zip(
                t.keys,
                t.values,
                t.operations
                if isinstance(t, ast.ConfigExpr)
                else [ast.ConfigEntryOperation.UNION] * len(t.keys),
            ):
                # Double star expression
                if not key:
                    value_vertex = Vertex.ast_to_vertex(
                        value, name=None, is_in_schema=is_in_schema
                    )
                    value_vertex.node = value
                    value_vertex.meta = Meta.from_ast_node(value)
                    Vertex.update_vertex_option(value_vertex, operation)
                    vertex.adjs.append(value_vertex)
                elif isinstance(key, ast.Identifier) and (
                    isinstance(t, ast.ConfigExpr) or is_in_schema
                ):
                    nest_key_len = len(key.names)
                    nest_vertex_list = [
                        Vertex(
                            name=key.names[i],
                            node=key,
                            adjs=[],
                            meta=Meta.from_ast_node(key),
                        )
                        for i in range(nest_key_len - 1)
                    ]
                    final_vertex = Vertex.ast_to_vertex(
                        value, name=key.names[-1], is_in_schema=is_in_schema
                    )
                    final_vertex.meta = Meta.from_ast_node(key)
                    if nest_key_len > 1:
                        # Link all vertex in nest vertex list
                        for i in range(nest_key_len - 2):
                            nest_vertex_list[i].adjs = [nest_vertex_list[i + 1]]
                        nest_vertex_list[-1].adjs = [final_vertex]
                    value_vertex = (
                        final_vertex if nest_key_len == 1 else nest_vertex_list[0]
                    )
                    value_vertex.meta = Meta.from_ast_node(key)
                    if isinstance(value, ast.SchemaExpr):
                        value_vertex.config_meta = ConfigMeta(
                            name=value.name.get_name(),
                            args=value.args,
                            kwargs=value.kwargs,
                        )
                    Vertex.update_vertex_option(
                        value_vertex, ast.ConfigEntryOperation.UNION
                    )
                    Vertex.update_vertex_option(final_vertex, operation)
                    vertex.adjs.append(value_vertex)
                else:
                    # Variable attributes that cannot be clearly expressed
                    # Such as string interpolation which denotes a runtime value
                    value_vertex = Vertex.ast_to_vertex(
                        value,
                        name=key.value
                        if isinstance(key, (ast.StringLit, ast.NumberLit))
                        else key,
                        is_in_schema=is_in_schema,
                    )
                    value_vertex.meta = Meta.from_ast_node(key)
                    Vertex.update_vertex_option(value_vertex, operation)
                    vertex.adjs.append(value_vertex)
            return vertex
        # Vertex end node and its adjs is None
        return Vertex(
            name=name,
            node=t,
            adjs=None,
            meta=Meta.from_ast_node(t),
        )

    @staticmethod
    def new_root(
        node: ast.Module = None, adjs: Optional[List["Vertex"]] = None
    ) -> "Vertex":
        """New a empty vertex root node"""
        return Vertex(
            name=VERTEX_ROOT_NAME,
            node=node,
            adjs=adjs,
            meta=Meta(
                filename=node.filename,
                line=1,
                column=1,
                end_line=1,
                end_column=1,
            ),
        )



"""The `ast` file contains the definitions of all KCL AST nodes
and operators and all AST nodes are derived from the `AST` class.
The main structure of a KCL program is as follows:

┌─────────────────────────────────────────────────────────────────┐
│                             Program                             │
│  ┌─────────────────┐  ┌─────────────────┐  ┌─────────────────┐  │
│  │   Main Package  │  │     Package1    │  │     Package2    │  │
│  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │
│  │  │  Module1  │  │  │  │  Module1  │  │  │  │  Module1  │  │  │
│  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │
│  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │
│  │  │  Module2  │  │  │  │  Module2  │  │  │  │  Module2  │  │  │
│  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │
│  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │
│  │  │    ...    │  │  │  │    ...    │  │  │  │    ...    │  │  │
│  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │
│  └─────────────────┘  └─────────────────┘  └─────────────────┘  │
└─────────────────────────────────────────────────────────────────┘

A single KCL file represents a module, which records file information,
package path information, and module document information, which is
mainly composed of all the statements in the KCL file.

The combination of multiple KCL files is regarded as a complete KCL
Program. For example, a single KCL file can be imported into KCL
files in other packages through statements such as import. Therefore,
the Program is composed of multiple modules, and each module is
associated with it. Corresponding to the package path.

:note: When the definition of any AST node is modified or the AST node
is added/deleted, it is necessary to modify the corresponding processing
in the compiler and regenerate the walker code.
:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""

import os
import json
import hashlib
import typing

from enum import Enum
from abc import ABC
from typing import List, Optional, Union, Dict
from pathlib import PosixPath

import kclvm.kcl.ast.lark_token as lark_token
import kclvm.kcl.ast as ast
import kclvm.internal.util.check_utils as check_utils

from .lark_token import TokenValue


class CmdArgSpec:
    """KCL command line argument spec, e.g. `kcl main.k -D name=value`"""

    def __init__(self, *, name: str = "", value: any = None):
        self.name: str = name
        self.value: any = value


class OverrideAction(Enum):
    CREATE_OR_UPDATE = "CreateOrUpdate"
    DELETE = "Delete"


class CmdOverrideSpec:
    """KCL command line override spec, e.g. `kcl main.k -O pkgpath:path.to.field=field_value`"""

    def __init__(
        self,
        *,
        pkgpath: str = "",
        field_path: str = "",
        field_value: str = "",
        action: OverrideAction = OverrideAction.CREATE_OR_UPDATE,
    ):
        self.pkgpath: str = pkgpath
        self.field_path: str = field_path
        self.field_value: str = field_value
        self.action: OverrideAction = action


class LarkToken(lark_token.LarkToken):
    @staticmethod
    def is_string(token: str):
        return token in [
            LarkToken.L_STRING,
            LarkToken.L_LONG_STRING,
        ]

    @staticmethod
    def is_int_number(token: str):
        return token in [
            LarkToken.L_DEC_NUMBER,
            LarkToken.L_HEX_NUMBER,
            LarkToken.L_BIN_NUMBER,
            LarkToken.L_OCT_NUMBER,
        ]

    @staticmethod
    def is_float_number(token: str):
        return token == LarkToken.L_FLOAT_NUMBER

    @staticmethod
    def is_name_constant(token: str):
        return token in [
            LarkToken.L_TRUE,
            LarkToken.L_FALSE,
            LarkToken.L_NONE,
            LarkToken.L_UNDEFINED,
        ]

    @staticmethod
    def get_token_value(token: str) -> str:
        return LarkToken.LL_token_str_value_map[token]


class BinOp(Enum):
    """BinOp is the set of all binary operators in KCL."""

    Add = LarkToken.L_PLUS
    Sub = LarkToken.L_MINUS
    Mul = LarkToken.L_MULTIPLY
    Div = LarkToken.L_DIVIDE
    Mod = LarkToken.L_MOD
    Pow = LarkToken.L_DOUBLE_STAR
    LShift = LarkToken.L_SHIFT_LEFT
    RShift = LarkToken.L_SHIFT_RIGHT
    BitOr = LarkToken.L_OR
    BitXor = LarkToken.L_XOR
    BitAnd = LarkToken.L_AND
    FloorDiv = LarkToken.L_DOUBLE_DIVIDE
    As = LarkToken.L_AS

    And = LarkToken.L_L_AND  # True and False
    Or = LarkToken.L_L_OR  # True or False

    @classmethod
    def enum_value_list(cls) -> [str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def enum_key_list(cls) -> [str]:
        return list(map(lambda c: c, cls))


class AugOp(Enum):
    Assign = LarkToken.L_ASSIGN
    Add = LarkToken.L_COMP_PLUS
    Sub = LarkToken.L_COMP_MINUS
    Mul = LarkToken.L_COMP_MULTIPLY
    Div = LarkToken.L_COMP_DIVIDE
    Mod = LarkToken.L_COMP_MOD
    Pow = LarkToken.L_COMP_DOUBLE_STAR
    LShift = LarkToken.L_COMP_SHIFT_LEFT
    RShift = LarkToken.L_COMP_SHIFT_RIGHT
    BitOr = LarkToken.L_COMP_OR
    BitXor = LarkToken.L_COMP_XOR
    BitAnd = LarkToken.L_COMP_AND
    FloorDiv = LarkToken.L_COMP_DOUBLE_DIVIDE

    @classmethod
    def enum_value_list(cls) -> [str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def enum_key_list(cls) -> [str]:
        return list(map(lambda c: c, cls))


class UnaryOp(Enum):
    UAdd = LarkToken.L_PLUS
    USub = LarkToken.L_MINUS
    Invert = LarkToken.L_NOT
    Not = LarkToken.L_L_NOT

    @classmethod
    def enum_value_list(cls) -> [str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def enum_key_list(cls) -> [str]:
        return list(map(lambda c: c, cls))


def judge_compare_op(optype: str) -> bool:
    return optype in CmpOp.enum_value_list()


CMP_OP_VALUE_LIST = [
    LarkToken.L_EQUAL_TO,
    LarkToken.L_NOT_EQUAL_TO,
    LarkToken.L_LESS_THAN,
    LarkToken.L_LESS_THAN_OR_EQUAL_TO,
    LarkToken.L_GREATER_THAN,
    LarkToken.L_GREATER_THAN_OR_EQUAL_TO,
    LarkToken.L_IS,
    LarkToken.L_IN,
    LarkToken.L_L_NOT,
]


class CmpOp(Enum):
    Eq = LarkToken.L_EQUAL_TO
    NotEq = LarkToken.L_NOT_EQUAL_TO
    Lt = LarkToken.L_LESS_THAN
    LtE = LarkToken.L_LESS_THAN_OR_EQUAL_TO
    Gt = LarkToken.L_GREATER_THAN
    GtE = LarkToken.L_GREATER_THAN_OR_EQUAL_TO
    Is = LarkToken.L_IS
    In = LarkToken.L_IN

    Not = LarkToken.L_L_NOT

    IsNot = f"{LarkToken.L_IS} {LarkToken.L_L_NOT}"  # "IS NOT"
    NotIn = f"{LarkToken.L_L_NOT} {LarkToken.L_IN}"  # "NOT IN"

    @classmethod
    def enum_value_list(cls) -> [str]:
        return CMP_OP_VALUE_LIST

    @classmethod
    def enum_key_list(cls) -> [str]:
        return list(map(lambda c: c, cls))


class ExprContext(Enum):
    LOAD = "LOAD"
    STORE = "STORE"
    DEL = "DEL"
    AUGLOAD = "AUGLOAD"
    AUGSTORE = "AUGSTORE"
    PARAM = "PARAM"

    @classmethod
    def enum_value_list(cls) -> [str]:
        return list(map(lambda c: c.value, cls))


AST_ENUM_LIST = {
    "BinOp": BinOp,
    "AugOp": AugOp,
    "UnaryOp": UnaryOp,
    "CmpOp": CmpOp,
    "ExprContext": ExprContext,
    "OverrideAction": OverrideAction,
}

OPERATOR_VALUE_MAP = {
    AugOp.Assign: TokenValue.ASSIGN,
    AugOp.Add: TokenValue.COMP_PLUS,
    AugOp.Sub: TokenValue.COMP_MINUS,
    AugOp.Mul: TokenValue.COMP_MULTIPLY,
    AugOp.Div: TokenValue.COMP_DIVIDE,
    AugOp.Mod: TokenValue.COMP_MOD,
    AugOp.Pow: TokenValue.COMP_DOUBLE_STAR,
    AugOp.LShift: TokenValue.COMP_SHIFT_LEFT,
    AugOp.RShift: TokenValue.COMP_SHIFT_RIGHT,
    AugOp.BitOr: TokenValue.COMP_OR,
    AugOp.BitXor: TokenValue.COMP_XOR,
    AugOp.BitAnd: TokenValue.COMP_AND,
    AugOp.FloorDiv: TokenValue.COMP_DOUBLE_DIVIDE,
    BinOp.Add: TokenValue.PLUS,
    BinOp.Sub: TokenValue.MINUS,
    BinOp.Mul: TokenValue.MULTIPLY,
    BinOp.Div: TokenValue.DIVIDE,
    BinOp.Mod: TokenValue.MOD,
    BinOp.Pow: TokenValue.DOUBLE_STAR,
    BinOp.LShift: TokenValue.SHIFT_LEFT,
    BinOp.RShift: TokenValue.SHIFT_RIGHT,
    BinOp.BitOr: TokenValue.OR,
    BinOp.BitXor: TokenValue.XOR,
    BinOp.BitAnd: TokenValue.AND,
    BinOp.FloorDiv: TokenValue.DOUBLE_DIVIDE,
    BinOp.And: TokenValue.L_AND,
    BinOp.Or: TokenValue.L_OR,
    BinOp.As: TokenValue.AS,
    CmpOp.Eq: TokenValue.EQUAL_TO,
    CmpOp.NotEq: TokenValue.NOT_EQUAL_TO,
    CmpOp.Lt: TokenValue.LESS_THAN,
    CmpOp.LtE: TokenValue.LESS_THAN_OR_EQUAL_TO,
    CmpOp.Gt: TokenValue.GREATER_THAN,
    CmpOp.GtE: TokenValue.GREATER_THAN_OR_EQUAL_TO,
    CmpOp.Is: TokenValue.IS,
    CmpOp.In: TokenValue.IN,
    CmpOp.Not: TokenValue.NOT,
    CmpOp.IsNot: " ".join([TokenValue.IS, TokenValue.L_NOT]),
    CmpOp.NotIn: " ".join([TokenValue.L_NOT, TokenValue.IN]),
    UnaryOp.UAdd: TokenValue.PLUS,
    UnaryOp.USub: TokenValue.MINUS,
    UnaryOp.Invert: TokenValue.NOT,
    UnaryOp.Not: TokenValue.L_NOT,
}


class Position:
    """Position describes an arbitrary source position including the filename,
    line, and column location.

    A Position is valid if the line number is > 0.
    The line and column are both 1 based.
    """

    def __init__(self, filename: str = None, line: int = None, column: int = None):
        self.filename: str = filename
        self.line: int = line
        self.column: int = column

    def is_valid(self) -> bool:
        return self.filename is not None and self.line is not None and self.line > 0

    def less(self, pos: "Position") -> bool:
        if not self.is_valid() or not pos or not pos.is_valid():
            return False
        if self.filename != pos.filename:
            return False
        if self.line < pos.line:
            return True
        if self.line == pos.line:
            return self.column < pos.column
        return False

    def less_equal(self, pos: "Position") -> bool:
        if not self.is_valid() or not pos or not pos.is_valid():
            return False
        if self.less(pos):
            return True
        return self == pos

    def __eq__(self, other: "Position") -> bool:
        return (
            self.filename == other.filename
            and self.line == other.line
            and self.column == other.column
        )

    def __str__(self) -> str:
        return f"<{self.filename}, ({self.line}, {self.column})>"


class AST:
    """
    All KCL node types implement the KCL AST interface
    """

    _line_offset: int = 0
    _column_offset: int = 0

    def __init__(
        self,
        line: Optional[int] = 0,
        column: Optional[int] = 0,
        end_line: Optional[int] = 0,
        end_column: Optional[int] = 0,
    ) -> None:
        self.filename: str = None
        self.relative_filename: str = None
        self.line: int = line
        self.column: int = column
        self.end_line: int = end_line
        self.end_column: int = end_column
        self.parent: Optional[AST] = None

    def __str__(self) -> str:
        return f"<{self.type}, ({self.line}, {self.column})>"

    def __repr__(self) -> str:
        return self.__str__()

    def get_line(self) -> int:
        """
        Get the node line, which is 1 based
        """
        return self.line

    def get_column(self) -> int:
        """
        Get the node column, which is 1 based
        """
        return self.column

    def get_end_line(self) -> int:
        """
        Get the node end_line
        """
        return self.end_line

    def get_end_column(self) -> int:
        """
        Get the node end_column
        """
        return self.end_column

    def set_ast_position(self, node, filename=None):
        import kclvm.compiler.parser.lark_pb2 as lark_pb

        check_utils.check_type_not_none(node, AST, lark_pb.Tree)
        check_utils.check_type_allow_none(filename, str, PosixPath)
        self.filename = filename if isinstance(node, lark_pb.Tree) else node.filename
        self.line = node.line + self._line_offset
        self.column = node.column + self._column_offset
        self.end_line = node.end_line + self._line_offset
        self.end_column = node.end_column + self._column_offset

        return self

    def set_column(self, column: int):
        assert isinstance(column, int) and column >= 0
        self.column = column

    def set_line(self, line: int):
        assert isinstance(line, int) and line >= 0
        self.line = line

    def set_end_line_column(self, ast_node):
        assert ast_node and isinstance(ast_node, AST)
        self.end_line = ast_node.end_line
        self.end_column = ast_node.end_column

    def set_end_line(self, end_line: int):
        assert isinstance(end_line, int) and end_line >= 0
        self.end_line = end_line

    def set_end_column(self, end_column: int):
        assert isinstance(end_column, int) and end_column >= 0
        self.end_column = end_column

    def offset_column(self, offset_column: int):
        assert isinstance(offset_column, int)
        self.column += offset_column

    def offset_line(self, offset_line: int):
        assert isinstance(offset_line, int)
        self.line += offset_line

    def offset_end_line(self, offset_end_line: int):
        assert isinstance(offset_end_line, int)
        self.end_line += offset_end_line

    def offset_end_column(self, offset_end_column: int):
        assert isinstance(offset_end_column, int)
        self.end_column += offset_end_column

    @classmethod
    def set_offset(cls, line_offset: int = 0, column_offset=0):
        cls._line_offset = line_offset
        cls._column_offset = column_offset

    @classmethod
    def reset_offset(cls):
        cls._line_offset = 0
        cls._column_offset = 0

    @property
    def type(self) -> str:
        return self.__class__.__name__

    @property
    def pos(self) -> Position:
        return Position(filename=self.filename, line=self.line, column=self.column)

    @property
    def end_pos(self) -> Position:
        return Position(
            filename=self.filename, line=self.end_line, column=self.end_column
        )

    def to_json(self, indent=4, sort_keys=False):
        return json.dumps(
            self,
            default=lambda o: o.value
            if type(o) in AST_ENUM_LIST.values()
            else {k: v for k, v in o.__dict__.items() if k != "parent"},
            indent=indent,
            sort_keys=sort_keys,
        )

    def contains_pos(self, pos: Position) -> bool:
        """
        check if current node contains a position
        :param pos: the given position
        :return: if current node contains the given position
        """
        start_pos = Position(
            filename=self.filename,
            line=self.line,
            column=self.column,
        )
        end_pos = Position(
            filename=self.filename,
            line=self.end_line,
            column=self.end_column,
        )
        return start_pos.less_equal(pos) and pos.less_equal(end_pos)

    def get_children(self) -> List["AST"]:
        def walk_field(inner: typing.Union[typing.List, typing.Dict, AST]):
            if isinstance(inner, list):
                [walk_field(item) for item in inner]
                return
            if isinstance(inner, dict):
                [walk_field(v) for _, v in inner]
                return
            if isinstance(inner, AST):
                children.append(inner)

        children = []
        [walk_field(field) for _, field in ast.iter_fields(self)]
        return children

    def find_leaf_by_pos(self, pos: Position) -> Optional["AST"]:
        if pos and pos.is_valid() and self.contains_pos(pos):
            children = self.get_children()
            if len(children) == 0:
                return self
            else:
                for child in children:
                    leaf = child.find_leaf_by_pos(pos)
                    if leaf:
                        return leaf
        return None

    def find_nearest_parent_by_type(self, tpe: typing.Type["AST"]) -> Optional["AST"]:
        parent = self.parent
        while parent:
            if parent.type == tpe.__name__:
                return typing.cast(tpe, parent)
            parent = parent.parent
        return None


class Stmt(AST):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Stmt"


class Expr(AST):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Expr"


class Name(Expr, ABC):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value: Optional[str] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Name"
        self.value: str = value


class TypeAliasStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "TypeAliasStmt"
        self.type_name: Optional[Identifier] = None
        self.type_value: Optional[Type] = None


class ExprStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ExprStmt"
        self.exprs: List[Expr] = []


class UnificationStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "UnificationStmt"
        self.target: Identifier = None
        self.value: SchemaExpr = None


class AssignStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "AssignStmt"
        self.targets: List[Identifier] = []
        self.value: Optional[Expr] = None
        self.type_annotation: str = ""
        self.type_annotation_node: Optional[Type] = None

    def __str__(self):
        return super().__str__()[:-1] + f" targets: {self.targets} value: {self.value}>"


class AugAssignStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "AugAssignStmt"
        self.op: Optional[AugOp] = None
        self.target: Optional[Identifier] = None
        self.value: Optional[Expr] = None

    def __str__(self):
        return (
            super().__str__()[:-1]
            + f" target: {self.target} augop: {self.op} value: {self.value}>"
        )


class AssertStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "AssertStmt"
        self.test: Optional[Expr] = None
        self.if_cond: Optional[Expr] = None
        self.msg: Optional[Expr] = None


class IfStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "IfStmt"
        self.cond: Optional[Expr] = None
        self.body: List[Stmt] = []
        self.elif_cond: List[Expr] = []
        self.elif_body: List[List[Stmt]] = []
        self.else_body: List[Stmt] = []

    def __str__(self):
        return (
            super().__str__()[:-1]
            + f" cond: {self.cond} body: {self.body} elif_cond: {self.elif_cond} elifbody: {self.elif_body} elsebody: {self.else_body}>"
        )


class ImportStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ImportStmt"
        self.path: Optional[str] = None
        self.name: Optional[str] = None
        self.asname: Optional[str] = None
        self.path_nodes: [Name] = []
        self.as_name_node: Optional[Name] = None

        self.rawpath: Optional[str] = None  # only for error message

    def __str__(self):
        return super().__str__()[:-1] + f" name: {self.name}>"

    @property
    def pkg_name(self) -> str:
        return self.asname or self.name


class SchemaIndexSignature(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "SchemaIndexSignature"
        self.key_name: Optional[str] = None
        self.key_type: Optional[str] = "str"
        self.value_type: Optional[str] = ""
        self.value: Optional[Expr] = None
        self.any_other: bool = False
        self.name_node: Optional[Name] = None
        self.value_type_node: Optional[Type] = None


class SchemaAttr(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "SchemaAttr"
        self.doc: str = ""
        self.name: str = ""
        self.type_str: str = ""
        self.op: Optional[Union[BinOp, AugOp]] = None
        self.value: Optional[Expr] = None
        self.is_optional: bool = False
        self.decorators: List[Decorator] = []
        self.name_node: Optional[Name] = None
        self.type_node: Optional[Type] = None

    def __str__(self):
        return (
            super().__str__()[:-1]
            + f" name: {self.name} type: {self.type_str} value: {self.value}>"
        )


class SchemaStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "SchemaStmt"
        self.doc: str = ""
        self.name: str = ""
        self.parent_name: Optional[Identifier] = None
        self.for_host_name: Optional[Identifier] = None
        self.is_mixin: bool = False
        self.is_protocol: bool = False
        self.args: Optional[Arguments] = None
        self.mixins: List[Identifier] = []
        self.body: List[Union[SchemaAttr, Stmt]] = []
        self.decorators: List[Decorator] = []
        self.checks: List[CheckExpr] = []
        self.index_signature: Optional[SchemaIndexSignature] = None
        self.name_node: Optional[Name] = None

    def __str__(self):
        return super().__str__()[:-1] + f" name: {self.name} body: {self.body}>"

    def has_only_attribute_definitions(self) -> bool:
        return not bool(
            self.args
            or self.mixins
            or self.checks
            or self.index_signature
            or any(
                [not isinstance(p, (SchemaAttr, UnificationStmt)) for p in self.body]
            )
        )

    def GetAttrList(self) -> List[SchemaAttr]:
        attr_list = []

        for attr in self.body or []:
            if isinstance(attr, (SchemaAttr, UnificationStmt)):
                attr_list.append(attr)

        return attr_list

    def GetAttrNameList(self) -> List[str]:
        attr_list = self.GetAttrList()
        return [attr.name for attr in attr_list]

    def GetIndexSignatureAttrName(self) -> Optional[str]:
        return self.index_signature.key_name if self.index_signature else None

    def GetStmtList(self) -> List[Stmt]:
        stmt_list = []

        for attr in self.body:
            if not isinstance(attr, SchemaAttr):
                stmt_list.append(attr)

        return stmt_list

    def GetLeftIdentifierList(self):
        """Get schema full attribute list including
        un-exported attributes and relaxed attributes
        """
        attr_list = []

        def loop_body(body: List[Stmt]):
            """Get the l-values recursively and add them into schema attr list"""
            if not body:
                return
            for stmt in body:
                if isinstance(stmt, AssignStmt):
                    for target in stmt.targets:
                        add_name(target.get_first_name())
                elif isinstance(stmt, (AugAssignStmt, UnificationStmt)):
                    add_name(stmt.target.get_first_name())
                elif isinstance(stmt, IfStmt):
                    loop_body(stmt.body)
                    for body in stmt.elif_body:
                        loop_body(body)
                    loop_body(stmt.else_body)
                elif isinstance(stmt, SchemaAttr):
                    add_name(stmt.name)

        def add_name(name: str):
            """Add the `name` into schema attr list"""
            if not name or not isinstance(name, str):
                return
            attr_list.append(name)

        loop_body(self.body)
        return attr_list


class RuleStmt(Stmt):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "RuleStmt"
        self.doc: str = ""
        self.name: str = ""
        self.parent_rules: List[Identifier] = []
        self.decorators: List[Decorator] = []
        self.checks: List[CheckExpr] = []
        self.name_node: Optional[Name] = None
        self.args: Optional[Arguments] = None
        self.for_host_name: Optional[Identifier] = None


class IfExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "IfExpr"

        # body if cond else orelse
        self.body: Optional[Expr] = None  # self.body
        self.cond: Optional[Expr] = None  # self.cond
        self.orelse: Optional[Expr] = None  # self.orelse


class UnaryExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "UnaryExpr"
        self.op: Optional[UnaryOp] = None
        self.operand: Optional[Expr] = None


class BinaryExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "BinaryExpr"
        self.left: Optional[Expr] = None
        self.op: Optional[Union[BinOp, CmpOp]] = None
        self.right: Optional[Expr] = None

    def __str__(self):
        return (
            super().__str__()[:-1]
            + f" left: {self.left} op: {self.op} right: {self.right}>"
        )


class SelectorExpr(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        ctx: ExprContext = ExprContext.LOAD,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "SelectorExpr"
        self.value: Optional[Expr] = None
        self.attr: Optional[Identifier] = None
        self.ctx: Optional[ExprContext] = ctx
        self.has_question: bool = False


class CallExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "CallExpr"
        self.func: Optional[Expr] = None
        self.args: List[Expr] = []
        self.keywords: List[Keyword] = []

    def __str__(self):
        return super().__str__()[:-1] + f" func: {self.func} args: {self.args}>"


class ParenExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ParenExpr"
        self.expr: Optional[Expr] = None


class QuantExpr(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        ctx: ExprContext = ExprContext.LOAD,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "QuantExpr"
        self.target: Optional[Expr] = None
        self.variables: List[Identifier] = []
        self.op: Optional[int] = None
        self.test: Optional[Expr] = None
        self.if_cond: Optional[Expr] = None
        self.ctx: ExprContext = ctx


class QuantOperation:
    ALL = 1
    ANY = 2
    FILTER = 3
    MAP = 4

    MAPPING = {
        "any": ANY,
        "all": ALL,
        "map": MAP,
        "filter": FILTER,
    }


class ListExpr(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        ctx: ExprContext = ExprContext.LOAD,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ListExpr"
        self.elts: List[Expr] = []
        self.ctx: ExprContext = ctx

    def __str__(self):
        return super().__str__()[:-1] + f" values: {self.elts}>"


class ListIfItemExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ListIfItemExpr"
        self.if_cond: Optional[Expr] = None
        self.exprs: List[Expr] = []
        self.orelse: Optional[Expr] = None


class ListComp(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ListComp"
        self.elt: Optional[Expr] = None
        self.generators: List[CompClause] = []


class StarredExpr(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        ctx: ExprContext = ExprContext.LOAD,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "StarredExpr"
        self.value: Optional[Expr] = None
        self.ctx: ExprContext = ctx


class DictComp(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "DictComp"
        self.key: Optional[Expr] = None
        self.value: Optional[Expr] = None
        self.operation: Optional[ConfigEntryOperation] = None
        self.generators: List[CompClause] = []


class ConfigIfEntryExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ConfigIfEntryExpr"
        self.if_cond: Optional[Expr] = None
        self.keys: List[Expr] = []
        self.values: List[Expr] = []
        self.operations: List[int] = []
        self.orelse: Optional[Expr] = None


class CompClause(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "CompClause"
        self.targets: List[Identifier] = []
        self.iter: Optional[Expr] = None
        self.ifs: List[Expr] = []


class SchemaExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "SchemaExpr"
        self.name: Optional[Identifier] = None
        self.args: List[Expr] = []
        self.kwargs: List[Keyword] = []
        self.config: Optional[ConfigExpr] = None

    def __str__(self):
        return super().__str__()[:-1] + f" name: {self.name} config: {self.config}>"


class ConfigExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ConfigExpr"
        self.items: List[ConfigEntry] = []

    @property
    def keys(self) -> List[Expr]:
        return [item.key for item in self.items]

    @property
    def values(self) -> List[Expr]:
        return [item.value for item in self.items]

    @property
    def operations(self) -> List[int]:
        return [item.operation for item in self.items]

    def __str__(self):
        return super().__str__()[:-1] + f" keys: {self.keys} values: {self.values}>"


class ConfigEntryOperation:
    UNION = 0
    OVERRIDE = 1
    INSERT = 2
    UNIQUE = 3
    UNIFICATION = 4
    MAPPING = {
        LarkToken.L_COLON: UNION,
        LarkToken.L_ASSIGN: OVERRIDE,
        LarkToken.L_COMP_PLUS: INSERT,
    }

    @staticmethod
    def get_min():
        return ConfigEntryOperation.UNION

    @staticmethod
    def get_max():
        return ConfigEntryOperation.UNIFICATION


class ConfigEntry(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        key: Expr = None,
        value: Expr = None,
        operation: int = ConfigEntryOperation.UNION,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ConfigEntry"
        self.key: Expr = key
        self.value: Expr = value
        self.operation: int = operation
        self.insert_index: int = -1

    def __str__(self):
        return super().__str__()[:-1] + f" key: {self.key} value: {self.value}>"


class CheckExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "CheckExpr"
        self.test: Optional[Expr] = None
        self.if_cond: Optional[Expr] = None
        self.msg: Optional[Expr] = None


class LambdaExpr(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "LambdaExpr"
        self.args: Optional[Arguments] = None
        self.return_type_str: Optional[str] = None
        self.return_type_node: Optional[Type] = None
        self.body: List[Stmt] = []


class Decorator(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Decorator"
        self.name: Optional[Identifier] = None
        self.args: Optional[CallExpr] = None


class Subscript(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Subscript"
        self.value: Optional[Expr] = None
        self.index: Optional[Expr] = None
        self.lower: Optional[Expr] = None
        self.upper: Optional[Expr] = None
        self.step: Optional[Expr] = None
        self.ctx: ExprContext = ExprContext.LOAD
        self.has_question: bool = False


class Keyword(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Keyword"
        self.arg: Optional[Identifier] = None
        self.value: Optional[Expr] = None


class Arguments(Expr):
    def __init__(
        self, line: Optional[int] = None, column: Optional[int] = None
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Arguments"
        self.args: List[Identifier] = []  # arg0, arg1, ...
        self.defaults: List[Expr] = []
        self.type_annotation_list: List[str] = []
        self.type_annotation_node_list: List[Type] = []

    def __str__(self):
        return super().__str__()[:-1] + f" args: {self.args}>"

    def GetArgName(self, i: int) -> str:
        if 0 <= i < len(self.args):
            return self.args[i].get_name()
        return ""

    def GetArgType(self, i: int) -> str:
        if 0 <= i < len(self.type_annotation_list):
            return self.type_annotation_list[i]
        return ""

    def SetArgType(self, i: int, tpe: str):
        if 0 <= i < len(self.type_annotation_list):
            self.type_annotation_list[i] = tpe

    def GetArgDefault(self, i: int) -> Optional[str]:
        if 0 <= i < len(self.defaults):
            return (
                typing.cast(StringLit, self.defaults[i]).value
                if self.defaults[i]
                else None
            )
        return None


class Compare(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Compare"
        self.left: Optional[Expr] = None
        self.ops: List[CmpOp] = []
        self.comparators: List[Expr] = []


class Identifier(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        names: List[str] = None,
        ctx: ExprContext = ExprContext.LOAD,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Identifier"
        self.names: List[str] = names if names else []
        self.pkgpath: str = ""
        self.ctx: ExprContext = ctx
        self.name_nodes: List[Name] = []

    def set_filename(self, filename: str) -> Expr:
        self.filename = filename
        return self

    def get_name(self, wrapper=True):
        return ".".join(self.names) if wrapper else self.names[-1]

    def get_first_name(self):
        return self.names[0]

    def set_ctx(self, ctx: ExprContext):
        self.ctx = ctx

    def __str__(self):
        return super().__str__()[:-1] + f" name: {self.get_name()}>"


class Literal(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value=None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Literal"
        self.value = value

    def __str__(self):
        return super().__str__()[:-1] + f" value: {self.value}>"


class NumberLit(Literal):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value: Optional[int] = None,
    ) -> None:
        super().__init__(line, column, value)
        self._ast_type = "NumberLit"
        self.binary_suffix: Optional[str] = None


class NumberBinarySuffix:
    n = "n"
    u = "u"
    m = "m"
    k = "k"
    K = "K"
    M = "M"
    G = "G"
    T = "T"
    P = "P"
    Ki = "Ki"
    Mi = "Mi"
    Gi = "Gi"
    Ti = "Ti"
    Pi = "Pi"


class StringLit(Literal):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value: Optional[str] = None,
    ) -> None:
        super().__init__(line, column, value)
        self._ast_type = "StringLit"
        self.is_long_string: bool = False
        self.raw_value: Optional[str] = None


class NameConstantLit(Literal):
    """
    Name constant: True, False, None
    """

    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value: Optional[bool] = None,
    ) -> None:
        super().__init__(line, column, value)
        self._ast_type = "NameConstantLit"


class JoinedString(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "JoinedString"
        self.is_long_string: bool = False
        self.values: List[Union[StringLit, FormattedValue]] = []
        self.raw_value: Optional[str] = None


class FormattedValue(Expr):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        value: Optional[Expr] = None,
        format_spec: str = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "FormattedValue"
        self.is_long_string: bool = False
        self.value: Optional[Expr] = value
        self.format_spec: str = format_spec


class Comment(AST):
    def __init__(
        self,
        filename: Optional[str] = None,
        line: Optional[int] = None,
        column: Optional[int] = None,
        end_line: Optional[int] = None,
        end_column: Optional[int] = None,
        text: Optional[str] = None,
    ) -> None:
        super().__init__(line, column, end_line=end_line, end_column=end_column)
        self.filename = filename
        self._ast_type = "Comment"
        self.text: Optional[str] = text

    def __str__(self):
        return super().__str__()[:-1] + f" text: {self.text}>"


class CommentGroup(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "CommentGroup"
        self.comments: List[Comment] = []


class Type(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Type"
        # type_element: schema_type | basic_type | compound_type | literal_type
        self.type_elements: List[
            Union[Identifier, BasicType, ListType, DictType, LiteralType]
        ] = []
        self.plain_type_str: str = ""


class BasicType(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "BasicType"
        self.type_name: str = ""


class ListType(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "ListType"
        self.inner_type: Optional[Type] = None
        self.plain_type_str: str = ""


class DictType(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "DictType"
        self.key_type: Optional[Type] = None
        self.value_type: Optional[Type] = None
        self.plain_type_str: str = ""


class LiteralType(AST):
    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "LiteralType"
        self.plain_value: str = ""
        self.value_type: str = ""
        self.string_value: Optional[StringLit] = None
        self.number_value: Optional[NumberLit] = None


class Module(AST):
    """Module is an abstract syntax tree for a single KCL file."""

    def __init__(
        self,
        line: Optional[int] = None,
        column: Optional[int] = None,
        pkg: str = "",
        filename: str = "",
    ) -> None:
        super().__init__(line, column)
        self._ast_type = "Module"
        self.pkg: str = pkg
        self.filename: str = filename
        self.body: List[Stmt] = []

        self.doc: str = ""
        self.name = self.pkg  # {__main__} or same as {self.pkg}

        self.global_names: List[str] = []
        self.local_names: Dict[str, List[str]] = {}

        self.comments: List[
            Comment
        ] = []  # List of all comments in the source KCL module

    def __str__(self):
        return (
            super().__str__()[:-1]
            + f" pkg: {self.pkg} filename: {self.filename} body: {self.body}>"
        )

    def GetPkgpath(self, asname: str) -> str:
        for stmt in self.body or []:
            if isinstance(stmt, ImportStmt):
                import_spec = typing.cast(ImportStmt, stmt)
                if import_spec.asname == asname:
                    return import_spec.path
        return ""

    def GetImportList(self) -> List[ImportStmt]:
        import_list: List[ImportStmt] = []

        for stmt in self.body or []:
            if isinstance(stmt, ImportStmt):
                import_list.append(stmt)

        return import_list

    def GetSchemaList(self) -> List[SchemaStmt]:
        schema_list: List[SchemaStmt] = []

        for stmt in self.body or []:
            if isinstance(stmt, SchemaStmt):
                schema_list.append(stmt)

        return schema_list

    def GetSchemaAndRuleList(self) -> List[Union[SchemaStmt, RuleStmt]]:
        schema_rule_list: List[Union[SchemaStmt, RuleStmt]] = []

        for stmt in self.body or []:
            if isinstance(stmt, (SchemaStmt, RuleStmt)):
                schema_rule_list.append(stmt)

        return schema_rule_list

    def GetFileName(self, root: str = "") -> str:
        """# Construct the filename from the root path and relative file."""
        if root and self.relative_filename and self.relative_filename[0] == ".":
            filename = root + self.relative_filename[1:]
        else:
            filename = self.filename
        return filename

    def GetFirstExprInExprStmt(self) -> Expr:
        if self.body and len(self.body) > 0:
            if isinstance(self.body[0], ExprStmt) and len(self.body[0].exprs) > 0:
                return self.body[0].exprs[0]
        return None


class ASTFactory:
    @staticmethod
    def get_ast_module(node, pkg, filename, name) -> Module:
        check_utils.check_type_allow_none(filename, str, PosixPath)
        check_utils.check_allow_none(pkg, str)

        p = Module()
        p.line = 1
        p.column = 1
        p.pkg = pkg
        p.set_ast_position(node, filename)
        p.name = name
        return p

    @staticmethod
    def get_ast_configentry(key, value, operation, filename) -> ConfigEntry:
        check_utils.check_type_allow_none(key, Identifier, StringLit)
        check_utils.check_allow_none(value, Expr)
        check_utils.check_type_allow_none(filename, str, PosixPath)
        p = ConfigEntry(
            line=key.line if key else value.line,
            column=key.column if key else value.column,
            key=key,
            value=value,
            operation=operation,
        )
        p.filename = filename
        return p

    @staticmethod
    def get_ast_identifier(value: str) -> Identifier:
        check_utils.check_not_none(value, str)
        p = Identifier()
        p.names = value.split(".")
        return p

    @staticmethod
    def get_ast_literal(tpe: typing.Type[AST], line, column, value, filename):
        assert tpe
        check_utils.check_allow_none(line, int)
        check_utils.check_allow_none(column, int)
        check_utils.check_type_allow_none(filename, str, PosixPath)
        p = tpe(line, column, value)
        p.filename = filename
        return typing.cast(tpe, p)

    @staticmethod
    def get_ast(tpe: typing.Type[AST], node, filename):
        assert tpe
        check_utils.check_type_allow_none(filename, str, PosixPath)
        p = tpe().set_ast_position(node, filename)
        return typing.cast(tpe, p)

    @staticmethod
    def get_op(tpe: typing.Type[Enum], op_type: str):
        assert tpe
        check_utils.check_allow_none(op_type, str)
        p = tpe(op_type) if op_type else None
        return typing.cast(tpe, p)

    @staticmethod
    def get_ast_formatted_value(value, format_spec, filename):
        check_utils.check_type_allow_none(filename, str, PosixPath)
        check_utils.check_allow_none(value, Expr)
        check_utils.check_allow_none(format_spec, str)
        p = FormattedValue(value=value, format_spec=format_spec)
        p.set_ast_position(value, filename)
        return p


class Program:
    """Program is the AST collection of all files of the running KCL program."""

    MAIN_PKGPATH = "__main__"

    def __init__(
        self,
        *,
        root: str = "",
        main: str = "",
        pkgs: Dict[str, List[Module]] = None,
        cmd_args: List[CmdArgSpec] = None,
        cmd_overrides: List[CmdOverrideSpec] = None,
    ):
        self.root: str = root if root else ""
        self.main: str = main if main else ""
        self.pkgs: Dict[str, List[Module]] = pkgs if pkgs else {}

        self.cmd_args: List[CmdArgSpec] = cmd_args if cmd_args else []
        self.cmd_overrides: List[CmdOverrideSpec] = (
            cmd_overrides if cmd_overrides else []
        )

    def get_check_sum(self, root: str = "") -> str:
        """Get the AST program all file md5 sum"""
        check_sum = hashlib.md5()
        for modules in self.pkgs.values():
            for module in modules:
                if (
                    root
                    and module.relative_filename
                    and module.relative_filename[0] == "."
                ):
                    filename = root + module.relative_filename[1:]
                else:
                    filename = module.filename
                if os.path.isfile(filename):
                    # Encoding the filename into the checksum
                    check_sum.update(
                        (filename.replace(root, ".", 1) if root else filename).encode(
                            encoding="utf-8"
                        )
                    )
                    with open(filename, "rb") as f:
                        check_sum.update(f.read())
        return check_sum.hexdigest()

    def to_json(self, indent=4, sort_keys=False):
        return json.dumps(
            self,
            default=lambda o: o.value
            if type(o) in AST_ENUM_LIST.values()
            else o.__dict__,
            indent=indent,
            sort_keys=sort_keys,
        )



#! /usr/bin/env python3
import re as _re


def KMANGLED_replace(string: str, pattern: str, replace: str, count: int = 0):
    """
    Return the string obtained by replacing the leftmost non-overlapping occurrences
    of the pattern in string by the replacement.
    """
    return _re.sub(pattern, replace, string, count)


def KMANGLED_match(string: str, pattern: str):
    """
    Try to apply the pattern at the start of the string, returning a Match object, or None if no match was found.
    """
    return bool(_re.match(pattern, string))


def KMANGLED_compile(pattern: str):
    """
    Compile a regular expression pattern, returning a bool value denoting whether the pattern is valid
    """
    return bool(_re.compile(pattern))


def KMANGLED_findall(string: str, pattern: str):
    """
    Return a list of all non-overlapping matches in the string.
    """
    return _re.findall(pattern, string)


def KMANGLED_search(string: str, pattern: str):
    """
    Scan through string looking for a match to the pattern,
    returning a Match object, or None if no match was found.
    """
    return bool(_re.search(pattern, string))


def KMANGLED_split(string: str, pattern: str, maxsplit: int = 0):
    """
    Scan through string looking for a match to the pattern,
    returning a Match object, or None if no match was found.
    """
    return _re.split(pattern, string, maxsplit)



from typing import List, Dict

import kclvm.kcl.error as kcl_error
import kclvm.tools.docs.model_pb2 as model


class SchemaDocStringChecker:
    """KCL schema docstring verification

    Verify that the schema docstring is consistent with
    the schema definition

    Parameters
    ----------
    schema: model.SchemaDoc
        The schema AST node

    model: model.SchemaDoc
        The schema document model
    """

    ATTR_NOT_FOUND_IN_DOC = (
        "Missing schema attribute description in schema: '{}', attribute: '{}'"
    )
    ATTR_NOT_FOUND_IN_SCHEMA = (
        "Redundant schema attribute description in schema: '{}', attribute: '{}'"
    )
    ATTR_INFO_INCONSISTENT_TYPE = "Inconsistent schema attribute info: schema: '{}', attribute: '{}', type in docstring: '{}', actual type : '{}'"

    ATTR_INFO_INCONSISTENT_OPTIONAL = "Inconsistent schema attribute info: schema: '{}', attribute: '{}', is_optional in docstring: '{}', actual is_optional : '{}'"

    def __init__(self, source_code: model.SchemaDoc, doc_string: model.SchemaDoc):
        self._source_code = source_code
        self._doc_string = doc_string

    def _check_attribute_diff(
        self,
        attr_map: Dict[str, model.SchemaAttributeDoc],
        attr_doc_map: Dict[str, model.SchemaAttributeDoc],
    ):
        """Verify that the schema docstring Attributes section describes exactly the same attributes as the schema body defines:
        1. all the attributes defined in the schema body must have the corresponding description in the docstring Attributes section
        2. all the attributes described in the docstring Attributes section must have the corresponding definition in the schema body
        """

        attr_not_in_doc = set(attr_map) - set(attr_doc_map)
        attr_not_in_schema = set(attr_doc_map) - set(attr_map)
        if attr_not_in_schema:
            kcl_error.report_warning(
                err_type=kcl_error.ErrType.InvalidDocstring_TYPE,
                arg_msg=self.ATTR_NOT_FOUND_IN_SCHEMA.format(
                    self._source_code.name,
                    ",".join(attr_not_in_schema),
                ),
            )
        if attr_not_in_doc:
            kcl_error.report_warning(
                err_type=kcl_error.ErrType.InvalidDocstring_TYPE,
                arg_msg=self.ATTR_NOT_FOUND_IN_DOC.format(
                    self._source_code.name,
                    ",".join(attr_not_in_doc),
                ),
            )

    def _check_attribute_def(
        self,
        attr_map: Dict[str, model.SchemaAttributeDoc],
        attr_doc_map: Dict[str, model.SchemaAttributeDoc],
    ):
        """Verify that each attribute in the schema docstring Attributes section is consistent with the corresponding attribute in schema body.
        Following features of the attribute will be verified:
        1. the attribute's type
        2. the attribute's optional info
        todo: the default value is not checked, since the code representation can be various, it can only be checked semantically.
        """
        common_attr_list = set(attr_map) & set(attr_doc_map)
        for attr in common_attr_list:
            schema_attr = attr_map[attr]
            schema_attr_doc = attr_doc_map[attr]
            if schema_attr.type.type_str.replace(
                " ", ""
            ) != schema_attr_doc.type.type_str.replace(" ", ""):
                kcl_error.report_warning(
                    err_type=kcl_error.ErrType.InvalidDocstring_TYPE,
                    arg_msg=self.ATTR_INFO_INCONSISTENT_TYPE.format(
                        self._source_code.name,
                        attr,
                        schema_attr_doc.type.type_str,
                        schema_attr.type.type_str,
                    ),
                )
            if schema_attr.is_optional != schema_attr_doc.is_optional:
                kcl_error.report_warning(
                    err_type=kcl_error.ErrType.InvalidDocstring_TYPE,
                    arg_msg=self.ATTR_INFO_INCONSISTENT_OPTIONAL.format(
                        self._source_code.name,
                        attr,
                        schema_attr_doc.is_optional,
                        schema_attr.is_optional,
                    ),
                )

    def _check(
        self,
        schema_attr_list: List[model.SchemaAttributeDoc],
        attr_doc_list: List[model.SchemaAttributeDoc],
    ):
        if not schema_attr_list or not attr_doc_list:
            return
        # Grouped by the schema attribute name
        attr_map = {attr.name: attr for attr in schema_attr_list}
        attr_doc_map = {attr.name: attr for attr in attr_doc_list}
        # Check attribute diff between schema docstring and schema AST
        self._check_attribute_diff(attr_map, attr_doc_map)
        # Check attribute definition between schema docstring and schema AST
        self._check_attribute_def(attr_map, attr_doc_map)

    def check(self):
        if (
            self._source_code is None
            or self._doc_string is None
            or not isinstance(self._source_code, model.SchemaDoc)
            or not isinstance(self._doc_string, model.SchemaDoc)
        ):
            return
        schema_attr_list = self._source_code.attributes
        attr_doc_list = self._doc_string.attributes
        self._check(schema_attr_list, attr_doc_list)



import os
import pathlib
from abc import ABC, abstractmethod

import kclvm.tools.docs.factory as factory
import kclvm.tools.docs.formats as doc_formats
import kclvm.tools.docs.model_pb2 as model
import kclvm.tools.docs.utils as utils


class LinkResolver(ABC):
    """
    path resolver refills the path links in Type.type_str
    """

    def __init__(self, doc_name_formatter, schema_section_render):
        self.doc_name_formatter = doc_name_formatter
        self.schema_section_render = schema_section_render

    @abstractmethod
    def resolve(
        self,
        doc: model.ModuleDoc,
        root: str,
        locale: str,
        with_locale_suffix: bool = False,
    ) -> model.ModuleDoc:
        pass


class MarkdownLinkResolver(LinkResolver):
    def resolve(
        self,
        module: model.ModuleDoc,
        root: str,
        locale: str,
        with_locale_suffix: bool = False,
    ) -> model.ModuleDoc:
        def add_link(
            tpe: model.Type, current_module: model.ModuleDoc, root: str
        ) -> str:
            def to_md_anchor(original: str) -> str:
                return "-".join(original.lower().split())

            def resolve_relative_path(to_dir: str, from_dir: str) -> str:
                if to_dir == from_dir:
                    return ""
                common_path = os.path.commonpath([from_dir, to_dir])
                upper_count = len(from_dir.strip(os.sep).split(os.sep)) - len(
                    common_path.strip(os.sep).split(os.sep)
                )

                return (
                    f"..{os.sep}" * upper_count
                    + to_dir.replace(common_path, "", 1).lstrip(os.sep)
                    + ("" if to_dir == common_path else os.sep)
                )

            if tpe.type_category in [
                model.Type.TypeCategory.ANY,
                model.Type.TypeCategory.BUILTIN,
                model.Type.TypeCategory.LIT,
                model.Type.TypeCategory.NAMED_CONSTANT,
            ]:
                return tpe.type_str
            if tpe.type_category == model.Type.TypeCategory.SCHEMA:
                assert tpe.schema_type
                section_link = to_md_anchor(
                    self.schema_section_render(tpe.schema_type.name)
                )
                from_path = pathlib.Path(current_module.relative_path)
                to_path = pathlib.Path(tpe.schema_type.relative_path)

                file_link = ""
                if str(from_path) != str(to_path):
                    # defines in different file
                    relative_dir = resolve_relative_path(
                        to_dir=str(to_path.parent), from_dir=str(from_path.parent)
                    )
                    file_link = f"{relative_dir}{self.doc_name_formatter(utils.module_name(tpe.schema_type.relative_path), locale, doc_formats.KCLDocFormat.MARKDOWN, with_locale_suffix)}"
                    # remove .md file extension suffix
                    file_link = file_link[:-3]
                return f"[{tpe.type_str}]({file_link}#{section_link})"
            if tpe.type_category == model.Type.TypeCategory.UNION:
                assert tpe.union_type
                return " \\| ".join(
                    [
                        add_link(t, current_module=current_module, root=root)
                        for t in tpe.union_type.types
                    ]
                )
            if tpe.type_category == model.Type.TypeCategory.DICT:
                assert tpe.dict_type
                key_type_str = add_link(
                    tpe.dict_type.key_type, current_module=current_module, root=root
                )
                value_type_str = add_link(
                    tpe.dict_type.value_type, current_module=current_module, root=root
                )
                return (
                    f"{{{key_type_str}: {value_type_str}}}"
                    if value_type_str
                    else f"{{{key_type_str}:}}"
                )
            if tpe.type_category == model.Type.TypeCategory.LIST:
                assert tpe.list_type
                return f"[{add_link(tpe.list_type.item_type, current_module=current_module, root=root)}]"
            return tpe.type_str

        for schema in module.schemas:
            for attr in schema.attributes:
                attr.type.type_str = add_link(attr.type, module, root)
            if schema.base_schema:
                schema.base_schema.type_str = add_link(schema.base_schema, module, root)
        return module


class YamlLinkResolver(LinkResolver):
    def resolve(
        self,
        module: model.ModuleDoc,
        root: str,
        locale: str,
        with_locale_suffix: bool = False,
    ) -> model.ModuleDoc:
        return module


class JsonLinkResolver(LinkResolver):
    def resolve(
        self,
        module: model.ModuleDoc,
        root: str,
        locale: str,
        with_locale_suffix: bool = False,
    ) -> model.ModuleDoc:
        return module


factory = factory.PathResolverFactory()
factory.register_format(doc_formats.KCLDocFormat.MARKDOWN, MarkdownLinkResolver)
factory.register_format(doc_formats.KCLDocFormat.YAML, YamlLinkResolver)
factory.register_format(doc_formats.KCLDocFormat.JSON, JsonLinkResolver)



# Copyright 2021 The KCL Authors. All rights reserved.

from .kcl_error import (
    KCLError,
    KCLException,
    KCLSyntaxException,
    KCLRuntimeException,
    KCLCompileException,
    KCLTypeException,
    UnKnownDecoratorError,
    DeprecatedError,
    InvalidDecoratorTargetError,
    KCLNameError,
    TypeRuntimeError,
    EvaluationError,
    KCLTabError,
    InvalidSyntaxError,
    CompileError,
    CannotFindModule,
    print_kcl_error_message,
    print_kcl_warning_message,
    print_common_error_message,
    print_internal_error_message,
    report_exception,
    report_warning,
    get_exception,
    ErrType,
)

from .kcl_err_theme import KCLErrMsgTheme, KCLErrMsgThemeDefault

from .kcl_err_template import (
    ErrLevel,
    ErrFileMsg,
    FileCache,
)

from .kcl_err_msg import (
    ErrEwcode,
    KCL_ERR_MSG,
    INT_OVER_FLOW_MSG,
    FLOAT_OVER_FLOW_MSG,
    FLOAT_UNDER_FLOW_MSG,
    DEPRECATED_WARNING_MSG,
    SCHEMA_CHECK_FILE_MSG_ERR,
    SCHEMA_CHECK_FILE_MSG_COND,
    RECURSIVE_LOADING_MODULE_MSG,
    CANNOT_FIND_MODULE_MSG,
    UNIQUE_KEY_MSG,
    CANNOT_ADD_MEMBERS_MSG,
    MULTI_INHERIT_MSG,
    INDENTATION_ERROR_MSG,
    UNKNOWN_DECORATOR_MSG,
    INVALID_DECORATOR_TARGET_MSG,
    NAME_ERROR_MSG,
    INVALID_FORMAT_SPEC_MSG,
)

__all__ = [
    "ErrLevel",
    "report_exception",
    "report_warning",
    "ErrType",
    "get_exception",
    "INT_OVER_FLOW_MSG",
    "FLOAT_OVER_FLOW_MSG",
    "FLOAT_UNDER_FLOW_MSG",
    "DEPRECATED_WARNING_MSG",
    "SCHEMA_CHECK_FILE_MSG_ERR",
    "SCHEMA_CHECK_FILE_MSG_COND",
    "UNIQUE_KEY_MSG",
    "CANNOT_ADD_MEMBERS_MSG",
    "RECURSIVE_LOADING_MODULE_MSG",
    "CANNOT_FIND_MODULE_MSG",
    "MULTI_INHERIT_MSG",
    "INDENTATION_ERROR_MSG",
    "UNKNOWN_DECORATOR_MSG",
    "INVALID_DECORATOR_TARGET_MSG",
    "NAME_ERROR_MSG",
    "INVALID_FORMAT_SPEC_MSG",
    "ErrFileMsg",
    "ErrEwcode",
    "print_kcl_error_message",
    "print_kcl_warning_message",
    "print_common_error_message",
    "print_internal_error_message",
    "KCLException",
    "KCLSyntaxException",
    "KCLCompileException",
    "KCLRuntimeException",
    "KCLTypeException",
    "UnKnownDecoratorError",
    "InvalidDecoratorTargetError",
    "KCLNameError",
    "TypeRuntimeError",
    "EvaluationError",
    "KCLError",
    "DeprecatedError",
    "KCLTabError",
    "InvalidSyntaxError",
    "CompileError",
    "CannotFindModule",
    "KCLErrMsgTheme",
    "KCLErrMsgThemeDefault",
    "KCL_ERR_MSG",
    "FileCache",
]



# Units based on SI and ECI
# Ref: https://en.wikipedia.org/wiki/Binary_prefix
# Kubernetes use case: https://pkg.go.dev/k8s.io/apimachinery/pkg/api/resource#Quantity

from typing import Union


IEC_SUFFIX = "i"
EXPONENTS = {"n": -3, "u": -2, "m": -1, "K": 1, "k": 1, "M": 2, "G": 3, "T": 4, "P": 5}
NUMBER_MULTIPLIER_REGEX = r"^([1-9][0-9]{0,63})(E|P|T|G|M|K|k|m|u|n|Ei|Pi|Ti|Gi|Mi|Ki)$"


def cal_num(value: int, suffix: str) -> int:
    """
    Calculate number based on value and binary suffix.

    Supported suffixes:
    SI: n | u | m | k | K | M | G | T | P
    IEC: Ki | Mi | Gi | Ti | Pi

    Input:
    value: int.
    suffix: str.

    Returns:
    int

    Raises:
    ValueError on invalid or unknown suffix
    """

    if not isinstance(value, int):
        raise ValueError("Unsupported value type: {}".format(type(value)))

    if not suffix:
        return value

    base = 1000
    unit = suffix

    validate_unit(unit)

    if unit[-1] == "i":
        base = 1024
        unit = unit[:-1]

    exponent = EXPONENTS[unit]
    return value * (base ** exponent)


def to_quantity(quantity: Union[str, int]) -> int:
    """
    Parse and return number based on input quantity.

    Supported suffixes:
    SI: n | u | m | k | K | M | G | T | P
    IEC: Ki | Mi | Gi | Ti | Pi

    Input:
    quantity: str.

    Returns:
    int

    Raises:
    ValueError on invalid or unknown input
    """
    if not isinstance(quantity, (int, str)):
        raise ValueError("Unsupported quantity type: {}".format(type(quantity)))

    if isinstance(quantity, int):
        return quantity

    number = quantity
    suffix = None
    if len(quantity) >= 2 and quantity[-1] == IEC_SUFFIX:
        if quantity[-2] in EXPONENTS:
            number = quantity[:-2]
            suffix = quantity[-2:]
    elif len(quantity) >= 1 and quantity[-1] in EXPONENTS:
        number = quantity[:-1]
        suffix = quantity[-1:]

    if not number:
        raise ValueError("Number can't be empty")

    number = int(number)

    if suffix is None:
        return number

    validate_unit(suffix[0])

    if suffix.endswith(IEC_SUFFIX):
        base = 1024
    else:
        base = 1000

    exponent = EXPONENTS[suffix[0]]
    return number * (base ** exponent)


def validate_unit(unit: str) -> None:
    # IEC validate
    if not unit or not isinstance(unit, str) or len(unit) > 2:
        raise ValueError("Invalid suffix {}".format(unit))

    if unit in ["ni", "ui", "mi", "ki"]:
        raise ValueError("Invalid suffix {}".format(unit))

    if unit[0] not in EXPONENTS:
        raise ValueError("Invalid suffix {}".format(unit))



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import Optional, Union, List, Callable
from abc import ABCMeta, abstractmethod

import kclvm.kcl.info as kcl_info
import kclvm.kcl.error as kcl_error

from .common import (
    LIST_TYPE_NAME,
    DICT_TYPE_NAME,
    SCHEMA_TYPE_NAME,
)

SCHEMA_SETTINGS_ATTR_NAME = "__settings__"
SELECTOR_ERROR = "SelectorError"
SELECTOR_EMPTY_VAR_ERROR_MSG = "selector expression variable can't be empty"
SELECTOR_INVALID_VAR_TYPE_ERROR_MSG = (
    "invalid selector expression variable type, expected list, dict and Schema"
)
SELECTOR_INVALID_EXPR_ERROR_MSG = "invalid selector expression"
SELECTOR_INVALID_USE_ERROR_MSG = "{} variable can't be used with {} selector expression"
SELECTOR_INVALID_CONDITION_ERROR_MSG = "invalid selector expression lambda condition"


SelectorVar = Union[list, dict]


class Selector(metaclass=ABCMeta):
    """
    Selector expression interface
    """

    def __init__(self, data: SelectorVar) -> None:
        self.data: SelectorVar = data

    @abstractmethod
    def get_all(self) -> Optional[SelectorVar]:
        """
        Get all child items form self.data
        """
        pass

    @abstractmethod
    def get_by_index(self, index: int) -> Optional[SelectorVar]:
        """
        Get all child items by index
        """
        pass

    @abstractmethod
    def get_by_keys(self, keys: List[str]) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child key is in keys
        """
        pass

    @abstractmethod
    def get_item_by_condition(self, condition: Callable) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child meets the condition
        """
        pass

    def get_var(self) -> SelectorVar:
        """
        Get data variable
        """
        return self.data

    def select(
        self,
        is_all: bool = False,
        index: Optional[int] = None,
        keys: List[str] = None,
        condition: Callable = None,
    ) -> Optional[SelectorVar]:
        if is_all:
            return self.get_all()
        if index is not None:
            return self.get_by_index(index)
        if keys:
            return self.get_by_keys(keys)
        if condition and callable(condition):
            return self.get_item_by_condition(condition)
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_EXPR_ERROR_MSG,
        )

    @staticmethod
    def _handle_collection(
        result: Optional[Union[list, dict]]
    ) -> Optional[Union[list, dict]]:
        """
        If the result len is zero, return None, else return itself
        """
        return result if result else None


class ListSelector(Selector):
    """
    List selector expression inherited from Selector
    """

    def __init__(self, data: SelectorVar) -> None:
        super().__init__(data)

    def get_all(self) -> SelectorVar:
        """
        Get all child items form self.data
        """
        return self.get_var()

    def get_by_index(self, index: int) -> Optional[SelectorVar]:
        """
        Get all child items by index
        """
        var = self.get_var()
        return var[index] if -len(var) <= index < len(var) else None

    def get_by_keys(self, keys: List[str]) -> SelectorVar:
        """
        Get all child items form self.data when child key is in keys
        """
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_USE_ERROR_MSG.format(
                LIST_TYPE_NAME, DICT_TYPE_NAME
            ),
        )

    def get_item_by_condition(self, condition: Callable) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child meets the condition
        """
        var = self.get_var()
        if condition.__code__.co_argcount == 1:
            return self._handle_collection([v for v in var if condition(v)])
        elif condition.__code__.co_argcount == 2:
            return self._handle_collection(
                [v for i, v in enumerate(var) if condition(i, v)]
            )
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_CONDITION_ERROR_MSG,
        )


class DictSelector(Selector):
    """
    List selector expression inherited from Selector
    """

    def __init__(self, data: SelectorVar) -> None:
        super().__init__(data)

    def get_all(self) -> SelectorVar:
        """
        Get all child items form self.data
        """
        return self.get_var()

    def get_by_index(self, index: int) -> SelectorVar:
        """
        Get all child items by index
        """
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_USE_ERROR_MSG.format(
                DICT_TYPE_NAME, LIST_TYPE_NAME
            ),
        )

    def get_by_keys(self, keys: List[str]) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child key is in keys
        """
        var = self.get_var()
        return self._handle_collection({k: var[k] for k in var if k in keys})

    def get_item_by_condition(self, condition: Callable) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child meets the condition
        """
        var = self.get_var()
        if condition.__code__.co_argcount == 1:
            return self._handle_collection({k: var[k] for k in var if condition(k)})
        elif condition.__code__.co_argcount == 2:
            return self._handle_collection(
                {k: v for k, v in var.items() if condition(k, v)}
            )
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_CONDITION_ERROR_MSG,
        )


class SchemaSelector(Selector):
    """
    Schema selector expression inherited from Selector
    """

    def __init__(self, data: SelectorVar) -> None:
        super().__init__(data)

    def get_all(self) -> Optional[SelectorVar]:
        """
        Get all child items form self.data
        """
        var = self.get_var()
        return self._handle_collection(
            {
                kcl_info.demangle(k): var[k]
                for k in var
                if kcl_info.ismangled(k) and SCHEMA_SETTINGS_ATTR_NAME not in k
            }
        )

    def get_by_index(self, index: int) -> SelectorVar:
        """
        Get all child items by index
        """
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_USE_ERROR_MSG.format(
                SCHEMA_TYPE_NAME, LIST_TYPE_NAME
            ),
        )

    def get_by_keys(self, keys: List[str]) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child key is in keys
        """
        data = self.get_all()
        return self._handle_collection({k: data[k] for k in data if k in keys})

    def get_item_by_condition(self, condition: Callable) -> Optional[SelectorVar]:
        """
        Get all child items form self.data when child meets the condition
        """
        var = self.get_var()
        if condition.__code__.co_argcount == 1:
            return self._handle_collection(
                {
                    kcl_info.demangle(k): var[k]
                    for k in var
                    if kcl_info.ismangled(k) and condition(kcl_info.demangle(k))
                }
            )
        elif condition.__code__.co_argcount == 2:
            return self._handle_collection(
                {
                    kcl_info.demangle(k): v
                    for k, v in var.items()
                    if kcl_info.ismangled(k) and condition(kcl_info.demangle(k), v)
                }
            )
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_CONDITION_ERROR_MSG,
        )


class SelectorFactory:
    """
    Factory class to build selector
    """

    @staticmethod
    def get(var: SelectorVar) -> Selector:
        """
        Get selector class using 'var'
        """

        def is_kcl_schema(obj):
            from kclvm.api.object import KCLSchemaObject

            return isinstance(obj, KCLSchemaObject)

        if isinstance(var, list):
            return ListSelector(var)
        if isinstance(var, dict):
            return DictSelector(var)
        if is_kcl_schema(var):
            return SchemaSelector(var)
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_INVALID_VAR_TYPE_ERROR_MSG,
        )


def select(
    var: SelectorVar,
    is_all: bool = False,
    index: Optional[int] = None,
    keys: List[str] = None,
    condition: Callable = None,
):
    """
    Use the selector expression to filter out the child elements of 'var'
    and return their references
    """
    if var is None:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.SelectorError_TYPE,
            arg_msg=SELECTOR_EMPTY_VAR_ERROR_MSG,
        )
    return SelectorFactory.get(var).select(is_all, index, keys or [], condition)



# Copyright 2020 The KCL Authors. All rights reserved.

import kclvm.kcl.error as kcl_error
import kclvm.api.object as objpkg

from .type import Type


def type_convert(obj: objpkg.KCLObject, tpe: Type) -> objpkg.KCLObject:
    """The type of `obj` is converted to another type.

    Raise an runtime error occurs in the type conversion.
    """
    if not obj or not isinstance(obj, objpkg.KCLObject):
        raise ValueError("Invalid parameter obj, expected KCL object")
    if not tpe or not isinstance(tpe, Type):
        raise ValueError("Invalid parameter tpe, expected KCL type object")
    if isinstance(obj, (objpkg.KCLNoneObject, objpkg.KCLUndefinedObject)):
        return obj
    if isinstance(tpe, objpkg.KCLAnyTypeObject):
        return obj
    if isinstance(tpe, objpkg.KCLIntTypeObject) and isinstance(
        obj, (objpkg.KCLIntObject, objpkg.KCLFloatObject)
    ):
        return objpkg.KCLIntObject(int(obj.value))
    if isinstance(tpe, objpkg.KCLFloatTypeObject) and isinstance(
        obj, (objpkg.KCLIntObject, objpkg.KCLFloatObject)
    ):
        return objpkg.KCLFloatObject(float(obj.value))
    if isinstance(tpe, objpkg.KCLStringTypeObject) and isinstance(
        obj, objpkg.KCLStringObject
    ):
        return obj
    if isinstance(tpe, objpkg.KCLBoolTypeObject) and isinstance(
        obj, objpkg.KCLNameConstantObject
    ):
        return obj
    if isinstance(tpe, objpkg.KCLListTypeObject) and isinstance(
        obj, objpkg.KCLListObject
    ):
        return objpkg.KCLListObject(
            items=[type_convert(item, tpe.item_type) for item in obj.items]
        )
    if isinstance(tpe, objpkg.KCLDictTypeObject) and isinstance(
        obj, objpkg.KCLDictObject
    ):
        if isinstance(obj, objpkg.KCLSchemaConfigObject):
            return objpkg.KCLSchemaConfigObject(
                operation_map=obj.operation_map,
                insert_index_map=obj.insert_index_map,
                value={
                    k: type_convert(obj.value[k], tpe.value_type) for k in obj.value
                },
            )
        else:
            return objpkg.KCLDictObject(
                value={k: type_convert(obj.value[k], tpe.value_type) for k in obj.value}
            )
    if (
        isinstance(tpe, objpkg.KCLSchemaTypeObject)
        and isinstance(obj, objpkg.KCLSchemaObject)
        and obj.runtime_type == tpe.runtime_type
    ):
        return obj
    if isinstance(tpe, objpkg.KCLSchemaDefTypeObject):
        return type_convert(obj, tpe.schema_type)
    kcl_error.report_exception(
        err_type=kcl_error.ErrType.EvaluationError_TYPE,
        file_msgs=[kcl_error.ErrFileMsg(filename=None, line_no=None, col_no=None)],
        arg_msg=f"Cannot convert type '{obj.type_str()}' to '{tpe.type_str()}'",
    )



"""
The `kcl_err_theme` file mainly contains the color constants needed to highlight the KCL error message.

ThemeId (Enum): The color theme of highlight KCL error message, currently there is only one `DEFAULT`.

ColorOption (Enum): Fields that can be highlighted in the KCL error message.

KCLErrMsgTheme: An abstract class, which specifies the methods required by KCL error message theme classes.

KCLErrMsgThemeDefault: It is the default implementation class of KCLErrMsgTheme.

Default highlight：

    EWCODE: green
    ERROR: red
    WARNING: yellow
    FILE_NAME: blue
    LINE_COLUMN: cyan
    SYNTAX: dark yellow
    COMPLIE: dark green
    RUNTIME: dark blue
    MARK: {
        1: red
        2: purple
    },

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""
from enum import Enum, unique
from kclvm.internal.util import PreCheck, CheckRules
from abc import ABCMeta, abstractmethod


@unique
class ThemeId(Enum):
    DEFAULT = 0
    OTHER = 1


@unique
class ColorOption(Enum):
    EWCODE = 0
    ERROR = 1
    WARNING = 2
    FILE_NAME = 3
    LINE_COLUMN = 4
    MARK = 5
    SYNTAX = 6
    COMPLIE = 7
    RUNTIME = 8


class KCLErrMsgTheme(metaclass=ABCMeta):
    @abstractmethod
    def color_mark(self, mark: str, mark_level: int = 1):
        pass

    @abstractmethod
    def color(self, content: str, color_option: ColorOption):
        pass


class KCLErrMsgThemeDefault(KCLErrMsgTheme):
    def __init__(self):
        self.KCL_THEME: dict = {
            ColorOption.EWCODE: "\033[0;92m{}\033[0m",  # green
            ColorOption.ERROR: "\033[0;91m{}\033[0m",  # red
            ColorOption.WARNING: "\033[0;93m{}\033[0m",  # yellow
            ColorOption.FILE_NAME: "\033[0;94m{}\033[0m",  # blue
            ColorOption.LINE_COLUMN: "\033[0;96m{}\033[0m",  # cyan
            ColorOption.SYNTAX: "\033[0;33m{}\033[0m",  # dark yellow
            ColorOption.COMPLIE: "\033[0;32m{}\033[0m",  # dark green
            ColorOption.RUNTIME: "\033[0;34m{}\033[0m",  # dark blue
            ColorOption.MARK: {
                1: "\033[0;31m{}\033[0m",  # red
                2: "\033[0;35m{}\033[0m",  # purple
            },
        }

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "mark")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "mark_level")
    @PreCheck((lambda v: CheckRules.check_int_range_allow_none(v, 1, 3)), "mark_level")
    def color_mark(self, mark: str, mark_level: int = 1):
        return self.KCL_THEME[ColorOption.MARK][mark_level].format(mark)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "content")
    @PreCheck(
        (lambda v: CheckRules.check_type_not_none(v, ColorOption)), "color_option"
    )
    def color(self, content: str, color_option: ColorOption):
        return self.KCL_THEME[color_option].format(content)



# Copyright 2021 The KCL Authors. All rights reserved.

from dataclasses import dataclass
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import IntEnum
from collections import defaultdict
from typing import Tuple, List, Optional, cast

import kclvm.kcl.ast as ast
import kclvm.compiler.astutil as astutil

from .vertex import Vertex
from .unifier import Unifier, UnifierConfig


INVALID_STRATEGY_MSG = "invalid strategy {}"


class BaseMerger(ABC):
    def __init__(self, module: ast.Module):
        self.module: ast.Module = module

    @abstractmethod
    def merge(self) -> Tuple[Vertex, Vertex, ast.Module]:
        raise NotImplementedError


@dataclass
class UnifyMerger(BaseMerger):
    def __init__(self, module: ast.Module):
        super().__init__(module)

    def unify_ast_module(
        self, config: UnifierConfig = UnifierConfig()
    ) -> Tuple[Optional[Vertex], Optional[Vertex], ast.Module]:
        """Unify vertex according to the AST Module"""
        if not self.module:
            return None, None, self.module
        vertex = Vertex.ast_to_vertex(self.module)
        unify_vertex = Unifier(config).unify(vertex)
        merged_module = cast(ast.Module, unify_vertex.vertex_to_ast())
        return vertex, unify_vertex, merged_module

    def merge(self) -> Tuple[Vertex, Vertex, ast.Module]:
        """Merge the AST Module

        Input
        -----
            None

        Output
        ------
            vertex: Vertex formed by AST before merging
            unify_vertex: Vertex formed by AST after merging
            module: The merged AST
        """
        vertex, unify_vertex, merged_module = self.unify_ast_module()
        self.module = _deal_origin_ast_module(merged_module, self.module)
        return vertex, unify_vertex, self.module


@dataclass
class UniqueUnifyMerger(UnifyMerger):
    def __init__(self, module: ast.Module):
        super().__init__(module)

    def merge(self) -> Tuple[Vertex, Vertex, ast.Module]:
        """Merge the AST Module

        Input
        -----
            None

        Output
        ------
            vertex: Vertex formed by AST before merging
            unify_vertex: Vertex formed by AST after merging
            module: The merged AST
        """
        vertex, unify_vertex, merged_module = self.unify_ast_module(
            config=UnifierConfig(check_unique=True)
        )
        self.module = _deal_origin_ast_module(merged_module, self.module)
        return vertex, unify_vertex, self.module


@dataclass
class Overrider(UnifyMerger):
    def __init__(self, module: ast.Module):
        super().__init__(module)

    def merge(self) -> Tuple[Vertex, Vertex, ast.Module]:
        vertex, unify_vertex, merged_module = self.unify_ast_module(
            config=UnifierConfig(override=True)
        )
        self.module = _deal_origin_ast_module(merged_module, self.module)
        return vertex, unify_vertex, self.module


class MergeStrategy(IntEnum):
    UNION = 1  # `:`
    OVERRIDE = 2  # `=`
    UNIQUE = 3  # `!`


class MergeStrategyFactory:
    MAPPING = {
        MergeStrategy.UNION: UnifyMerger,
        MergeStrategy.OVERRIDE: Overrider,
        MergeStrategy.UNIQUE: UniqueUnifyMerger,
    }

    @staticmethod
    def get(strategy: MergeStrategy):
        return MergeStrategyFactory.MAPPING.get(strategy, UnifyMerger)


def _deal_origin_ast_module(
    merged_module: ast.Module, origin_module: ast.Module, all_visited: bool = False
) -> Optional[ast.Module]:
    """Deal origin AST Module according to merged_module"""
    if not merged_module or not isinstance(merged_module, ast.Module):
        return None
    merged_declaration_list = astutil.filter_declarations(merged_module, ast.SchemaExpr)
    merged_declaration_map = defaultdict(list)
    for d in merged_declaration_list:
        d.value.filename = d.filename
        merged_declaration_map[d.name].append(d.value)
    merged_declaration_visited = {d.name: False for d in merged_declaration_list}
    # Reverse traversal
    origin_module.body = origin_module.body[::-1]
    i = 0
    # TODO: Optimize the while loop using AST Transformer.
    while i < len(origin_module.body):
        stmt = origin_module.body[i]
        if isinstance(stmt, ast.UnificationStmt):
            target = stmt.target
            name = target.get_first_name()
            if name in merged_declaration_map:
                del origin_module.body[i]
                i -= 1
                values = merged_declaration_map[name]
                if values and not merged_declaration_visited[name] and not all_visited:
                    identifier = ast.Identifier(
                        line=target.line,
                        column=target.column,
                        names=[name],
                        ctx=ast.ExprContext.STORE,
                    )
                    identifier.pkgpath = target.pkgpath
                    identifier.filename = target.filename
                    identifier.end_line, identifier.end_column = (
                        target.end_line,
                        target.end_column,
                    )
                    insert_unification_stmt = ast.UnificationStmt(
                        line=stmt.line, column=stmt.column
                    )
                    insert_unification_stmt.target = identifier
                    (
                        insert_unification_stmt.end_line,
                        insert_unification_stmt.end_column,
                    ) = (
                        stmt.end_line,
                        stmt.end_column,
                    )
                    for value in values:
                        stmt_copy = deepcopy(insert_unification_stmt)
                        stmt_copy.value = value
                        stmt_copy.filename = value.filename
                        stmt_copy.line = value.line
                        stmt_copy.column = value.column
                        stmt_copy.end_line = value.end_line
                        stmt_copy.end_column = value.end_column
                        i += 1
                        origin_module.body.insert(i, stmt_copy)
                    merged_declaration_visited[name] = True
        elif isinstance(stmt, ast.AssignStmt) and isinstance(
            stmt.value, ast.SchemaExpr
        ):
            if not stmt.targets:
                del origin_module.body[i]
                i -= 1
            j = 0
            while j < len(stmt.targets):
                target = stmt.targets[j]
                name = target.get_first_name()
                if name in merged_declaration_map:
                    del stmt.targets[j]
                    j -= 1
                    if len(stmt.targets) == 0:
                        del origin_module.body[i]
                        i -= 1
                    values = merged_declaration_map[name]
                    if (
                        values
                        and not merged_declaration_visited[name]
                        and not all_visited
                    ):
                        identifier = ast.Identifier(
                            line=target.line,
                            column=target.column,
                            names=[name],
                            ctx=ast.ExprContext.STORE,
                        )
                        identifier.pkgpath = target.pkgpath
                        identifier.filename = target.filename
                        identifier.end_line, identifier.end_column = (
                            target.end_line,
                            target.end_column,
                        )
                        insert_assign_stmt = ast.AssignStmt(
                            line=stmt.line, column=stmt.column
                        )
                        insert_assign_stmt.targets = [identifier]
                        insert_assign_stmt.end_line, insert_assign_stmt.end_column = (
                            stmt.end_line,
                            stmt.end_column,
                        )
                        insert_assign_stmt.type_annotation = stmt.type_annotation
                        insert_assign_stmt.type_annotation_node = (
                            stmt.type_annotation_node
                        )
                        for value in values:
                            stmt_copy = deepcopy(insert_assign_stmt)
                            stmt_copy.value = value
                            stmt_copy.filename = value.filename
                            stmt_copy.line = value.line
                            stmt_copy.column = value.column
                            stmt_copy.end_line = value.end_line
                            stmt_copy.end_column = value.end_column
                            i += 1
                            origin_module.body.insert(i, stmt_copy)
                        merged_declaration_visited[name] = True
                j += 1
        i += 1
    # Remove empty targets assignment
    origin_module.body = [
        m
        for m in reversed(origin_module.body)
        if not isinstance(m, ast.AssignStmt) or m.targets
    ]
    return origin_module


def MergeASTList(
    modules: List[ast.Module], strategy: MergeStrategy = MergeStrategy.UNION
) -> List[ast.Module]:
    """Merge the configurations of the same name in the
    AST Module list, and the ones that cannot be merged
    will be handed over to the VM for calculation.
    """
    if not modules or not isinstance(modules, list):
        return []
    # AST module filename list
    filenames = [m.filename for m in modules]
    # Config need to be merged
    file_configs = [
        stmt
        for m in modules[1:]
        for stmt in m.body
        if (isinstance(stmt, ast.AssignStmt) and isinstance(stmt.value, ast.SchemaExpr))
        or isinstance(stmt, ast.UnificationStmt)
        or isinstance(stmt, ast.ImportStmt)
    ]
    # Config filename meta
    files_meta = [
        (m.filename, stmt.line, stmt.column, stmt.end_line, stmt.end_column)
        for m in modules[1:]
        for stmt in m.body
        if (isinstance(stmt, ast.AssignStmt) and isinstance(stmt.value, ast.SchemaExpr))
        or isinstance(stmt, ast.UnificationStmt)
        or isinstance(stmt, ast.ImportStmt)
    ]
    # Record the statement filename
    for i, config in enumerate(file_configs):
        (
            file_configs[i].filename,
            file_configs[i].line,
            file_configs[i].column,
            file_configs[i].end_line,
            file_configs[i].end_column,
        ) = files_meta[i]
    if not modules[0].body:
        modules[0].body = []
    modules[0].body += file_configs
    MergeAST(modules[0], strategy)
    # Other file config list
    file_configs = [
        stmt
        for stmt in modules[0].body
        if stmt.filename is not None
        and stmt.filename != modules[0].filename
        and not isinstance(stmt, ast.ImportStmt)
    ]
    # Origin module
    modules[0].body = [stmt for stmt in modules[0].body if stmt not in file_configs]
    # Filter all config except the first file
    for i, _ in enumerate(modules[1:]):
        modules[i + 1].body = [
            stmt
            for stmt in modules[i + 1].body
            if (
                not isinstance(stmt, ast.AssignStmt)
                or not isinstance(stmt.value, ast.SchemaExpr)
            )
            and not isinstance(stmt, ast.UnificationStmt)
        ]
    # Insert the merged configuration into different files
    for config in file_configs:
        index = filenames.index(config.filename)
        line = config.line
        insert_index = len(modules[index].body)
        for i, stmt in enumerate(modules[index].body):
            if stmt.line >= line:
                insert_index = i
                break
        modules[index].body.insert(insert_index, config)
    # Return the merged multi-file modules
    return modules


def MergeAST(
    module: ast.Module, strategy: MergeStrategy = MergeStrategy.UNION
) -> ast.Module:
    """Merge the configurations of the same name in the single
    AST module, and the ones that cannot be merged will be handed
    over to the VM for calculation.
    """
    _, _, merged_module = MergeASTToVertex(module, strategy)
    return merged_module


def MergeASTToVertex(
    module: ast.Module, strategy: MergeStrategy = MergeStrategy.UNION
) -> Tuple[Optional[Vertex], Optional[Vertex], ast.Module]:
    """Merge the configurations of the same name in the AST and
    return the merged vertices, and the ones that cannot be merged
    will be handed over to the VM for calculation.
    """
    if not module:
        return None, None, module
    merger = MergeStrategyFactory.get(strategy)
    return merger(module).merge()



# Copyright 2021 The KCL Authors. All rights reserved.

from .option import (
    kcl_option,
    kcl_option_exec,
    kcl_option_init_all,
    kcl_option_init,
    kcl_option_reset,
    kcl_option_check,
)
from .decorators import (
    decorator_factory,
    Decorator,
    DecoratorTargetType,
    Deprecated,
    Info,
)
from .selector import select
from .undefined import Undefined, UndefinedType
from .path_selector import (
    is_selector_mode,
    build_selector_index,
    select_instance_attributes,
)

__all__ = [
    "kcl_option",
    "kcl_option_exec",
    "kcl_option_init_all",
    "kcl_option_init",
    "kcl_option_reset",
    "kcl_option_check",
    "select",
    "is_selector_mode",
    "build_selector_index",
    "select_instance_attributes",
    "decorator_factory",
    "Decorator",
    "DecoratorTargetType",
    "Deprecated",
    "Info",
    "Undefined",
    "UndefinedType",
]



from typing import List

from kclvm.tools.lint.message.message import Message


class BaseReporter:
    name: str = ""
    """base class for reporters"""

    def __init__(self, linter, output=None, encoding=None):
        self.linter = linter
        self.out = None
        self.out_encoding = None
        self.set_output(output, encoding)

    def __eq__(self, other):
        return self.name == other.name and self.linter == other.linter

    def set_output(self, output=None, encoding="utf-8"):
        """
        set output stream
        todo:
        The output property is not used yet， and replaced by open_output_stream.
        When reporter display the result, the 'open_output_stream' function is
        called to open the output stream.
        """
        self.out = output
        self.out_encoding = encoding

    def display(self):
        self.print_msg(self.linter.msgs)

    def print_msg(self, msgs: List[Message] = None):
        """Should be overridden by subclass"""
        raise NotImplementedError()



import os
from io import StringIO
from typing import Union, Optional
import pathlib

import kclvm.kcl.ast.ast as ast
import kclvm.tools.printer.printer as printer
import kclvm.kcl.info as kcl_info


def get_source_code(
    file: Optional[Union[str, pathlib.PosixPath]], line: int, code: Optional[str] = None
) -> str:
    if code:
        lines = code.split("\n")
        assert line <= len(lines)
        source_line = lines[line - 1]
    else:
        _file = str(file)
        assert is_kcl_file(_file)
        assert line > 0
        with open(_file, "r", encoding="utf8") as source_file:
            lines = source_file.read().split("\n")
            assert line <= len(lines)
            source_line = lines[line - 1]
    return source_line


def is_kcl_file(file: str) -> bool:
    assert isinstance(file, str)
    return os.path.isfile(file) and file.endswith(kcl_info.KCL_FILE_SUFFIX)


def get_code_from_module(module: ast.Module) -> str:
    assert isinstance(module, ast.Module)
    with StringIO() as IO:
        printer.PrintAST(module, IO)
        code = IO.getvalue()
    return code



from .check_type import (
    check,
    check_type,
    check_type_dict,
    runtime_type,
    runtime_types,
    type_pack_and_check,
    check_type_builtin,
    has_literal_type,
)

__all__ = [
    "check",
    "check_type",
    "check_type_dict",
    "runtime_type",
    "runtime_types",
    "type_pack_and_check",
    "check_type_builtin",
    "has_literal_type",
]



from .printer import PrintAST, Config
from .splice import SchemaRuleCodeSnippet, splice_schema_with_rule

__all__ = [
    "PrintAST",
    "Config",
    "SchemaRuleCodeSnippet",
    "splice_schema_with_rule",
]



# -*- coding: utf-8 -*-

# Parser for protocol buffer .proto files
# Thanks to https://github.com/python-parsy/parsy/blob/master/examples/proto3.py
import enum as stdlib_enum
from string import ascii_letters, digits, hexdigits, octdigits

import attr

from parsy import char_from, from_enum, generate, regex, seq, string

# This file follows the spec at
# https://developers.google.com/protocol-buffers/docs/reference/proto3-spec
# very closely.

# However, because we are parsing into useful objects, we do transformations
# along the way e.g. turning into integers, strings etc. and custom objects.
# Some of the lowest level items have been implemented using 'regex' and converting
# the descriptions to regular expressions. Higher level constructs have been
# implemented using other parsy primitives and combinators.

# Notes:

# 1. Whitespace is very badly defined in the 'spec', so we guess what is meant.
# 2. The spec doesn't allow for comments, and neither does this parser.
#    Other places mention that C++ style comments are allowed. To support that,
#    this parser would need to be changed into split lexing/parsing stages
#    (otherwise you hit issues with comments start markers within string literals).
# 3. Other notes inline.


# Our utilities
def optional_string(s):
    return string(s).times(0, 1).concat()


def convert_octal(s):
    return int(s, 8)


def convert_hex(s):
    return int(s, 16)


def exclude_none(data):
    return [i for i in data if i is not None]


convert_decimal = int


def lexeme(p):
    """
    From a parser (or string), make a parser that consumes
    whitespace on either side.
    """
    if isinstance(p, str):
        p = string(p)
    return regex(r"\s*") >> p << regex(r"\s*")


def is_present(p):
    """
    Given a parser or string, make a parser that returns
    True if the parser matches, False otherwise
    """
    return lexeme(p).optional().map(lambda v: False if v is None else True)


# Our data structures
@attr.s
class Import:
    identifier = attr.ib()
    option = attr.ib()


@attr.s
class Package:
    identifier = attr.ib()


@attr.s
class Option:
    name = attr.ib()
    value = attr.ib()


@attr.s
class Field:
    repeated = attr.ib()
    type = attr.ib()
    name = attr.ib()
    number = attr.ib()
    options = attr.ib()


@attr.s
class OneOfField:
    type = attr.ib()
    name = attr.ib()
    number = attr.ib()
    options = attr.ib()


@attr.s
class OneOf:
    name = attr.ib()
    fields = attr.ib()


@attr.s
class Map:
    key_type = attr.ib()
    type = attr.ib()
    name = attr.ib()
    number = attr.ib()
    options = attr.ib()


@attr.s
class Reserved:
    items = attr.ib()


@attr.s
class Range:
    from_ = attr.ib()
    to = attr.ib()


@attr.s
class EnumField:
    name = attr.ib()
    value = attr.ib()
    options = attr.ib()


@attr.s
class Enum:
    name = attr.ib()
    body = attr.ib()


@attr.s
class Message:
    name = attr.ib()
    body = attr.ib()


@attr.s
class Service:
    name = attr.ib()
    body = attr.ib()


@attr.s
class Rpc:
    name = attr.ib()
    request_stream = attr.ib()
    request_message_type = attr.ib()
    response_stream = attr.ib()
    response_message_type = attr.ib()
    options = attr.ib()


@attr.s
class Proto:
    syntax = attr.ib()
    statements = attr.ib()


# Enums:
class ImportOption(stdlib_enum.Enum):
    WEAK = "weak"
    PUBLIC = "public"


class Type(stdlib_enum.Enum):
    DOUBLE = "double"
    FLOAT = "float"
    INT32 = "int32"
    INT64 = "int64"
    UINT32 = "uint32"
    UINT64 = "uint64"
    SINT32 = "sint32"
    SINT64 = "sint64"
    FIXED32 = "fixed32"
    FIXED64 = "fixed64"
    SFIXED32 = "sfixed32"
    SFIXED64 = "sfixed64"
    BOOL = "bool"
    STRING = "string"
    BYTES = "bytes"


class KeyType(stdlib_enum.Enum):
    INT32 = "int32"
    INT64 = "int64"
    UINT32 = "uint32"
    UINT64 = "uint64"
    SINT32 = "sint32"
    SINT64 = "sint64"
    FIXED32 = "fixed32"
    FIXED64 = "fixed64"
    SFIXED32 = "sfixed32"
    SFIXED64 = "sfixed64"
    BOOL = "bool"
    STRING = "string"


# Some extra constants to avoid typing
SEMI = lexeme(";")
EQ = lexeme("=")
LPAREN = lexeme("(")
RPAREN = lexeme(")")
LBRACE = lexeme("{")
RBRACE = lexeme("}")

# -- Beginning of following spec --
# Letters and digits
letter = char_from(ascii_letters)
decimalDigit = char_from(digits)
octalDigit = char_from(octdigits)
hexDigit = char_from(hexdigits)

# Identifiers

# Compared to spec, we add some '_' prefixed items which are not wrapped in `lexeme`,
# on the assumption that spaces in the middle of identifiers are not accepted.
_ident = (letter + (letter | decimalDigit | string("_")).many().concat()).desc("ident")
ident = lexeme(_ident)
fullIdent = lexeme(ident + (string(".") + ident).many().concat()).desc("fullIdent")
_messageName = _ident
messageName = lexeme(ident).desc("messageName")
_enumName = ident
enumName = lexeme(_enumName).desc("enumName")
fieldName = ident.desc("fieldName")
oneofName = ident.desc("oneofName")
mapName = ident.desc("mapName")
serviceName = ident.desc("serviceName")
rpcName = ident.desc("rpcName")
messageType = (
    optional_string(".") + (_ident + string(".")).many().concat() + _messageName
)
enumType = optional_string(".") + (_ident + string(".")).many().concat() + _enumName

# Integer literals
decimalLit = regex("[1-9][0-9]*").desc("decimalLit").map(convert_decimal)
octalLit = regex("0[0-7]*").desc("octalLit").map(convert_octal)
hexLit = regex("0[x|X][0-9a-fA-F]+").desc("octalLit").map(convert_hex)
intLit = decimalLit | octalLit | hexLit


# Floating-point literals
decimals = r"[0-9]+"
exponent = r"[e|E][+|-]?" + decimals
floatLit = (
    regex(
        r"({decimals}\.({decimals})?({exponent})?)|{decimals}{exponent}|\.{decimals}({exponent})?".format(
            decimals=decimals, exponent=exponent
        )
    )
    .desc("floatLit")
    .map(float)
)


# Boolean
boolLit = (string("true").result(True) | string("false").result(False)).desc("boolLit")


# String literals
hexEscape = regex(r"\\[x|X]") >> regex("[0-9a-fA-F]{2}").map(convert_hex).map(chr)
octEscape = regex(r"\\") >> regex("[0-7]{2}").map(convert_octal).map(chr)
charEscape = regex(r"\\") >> (
    string("a").result("\a")
    | string("b").result("\b")
    | string("f").result("\f")
    | string("n").result("\n")
    | string("r").result("\r")
    | string("t").result("\t")
    | string("v").result("\v")
    | string("\\").result("\\")
    | string("'").result("'")
    | string('"').result('"')
)
escapes = hexEscape | octEscape | charEscape
# Correction to spec regarding " and ' inside quoted strings
strLit = (
    string("'") >> (escapes | regex(r"[^\0\n\'\\]")).many().concat() << string("'")
    | string('"') >> (escapes | regex(r"[^\0\n\"\\]")).many().concat() << string('"')
).desc("strLit")
quote = string("'") | string('"')

# EmptyStatement
emptyStatement = string(";").result(None)


# Signed numbers:
def signedNumberChange(s, num):
    """(Extra compared to spec, to cope with need to produce signed numeric values)"""
    return (-1) if s == "-" else (+1)


sign = regex("[-+]?")
signedIntLit = seq(sign, intLit).combine(signedNumberChange)
signedFloatLit = seq(sign, floatLit).combine(signedNumberChange)


# Constant
# put fullIdent at end to disabmiguate from boolLit
constant = signedIntLit | signedFloatLit | strLit | boolLit | fullIdent

# Syntax
syntax = lexeme("syntax") >> EQ >> quote >> string("proto3") << quote + SEMI

# Import Statement
import_option = from_enum(ImportOption)

import_ = seq(
    lexeme("import") >> import_option.optional().tag("option"),
    lexeme(strLit).tag("identifier") << SEMI,
).combine_dict(Import)

# Package
package = seq(lexeme("package") >> fullIdent << SEMI).map(Package)

# Option
optionName = (ident | (LPAREN >> fullIdent << RPAREN)) + (
    string(".") + ident
).many().concat()
option = seq(
    lexeme("option") >> optionName.tag("name"),
    EQ >> constant.tag("value") << SEMI,
).combine_dict(Option)

# Normal field
type_ = lexeme(from_enum(Type) | messageType | enumType)
fieldNumber = lexeme(intLit)

fieldOption = seq(optionName.tag("name"), EQ >> constant.tag("value")).combine_dict(
    Option
)
fieldOptions = fieldOption.sep_by(lexeme(","), min=1)
fieldOptionList = (
    (lexeme("[") >> fieldOptions << lexeme("]"))
    .optional()
    .map(lambda o: [] if o is None else o)
)

field = seq(
    is_present("repeated").tag("repeated"),
    type_.tag("type"),
    fieldName.tag("name") << EQ,
    fieldNumber.tag("number"),
    fieldOptionList.tag("options") << SEMI,
).combine_dict(Field)


# Oneof and oneof field
oneofField = seq(
    type_.tag("type"),
    fieldName.tag("name") << EQ,
    fieldNumber.tag("number"),
    fieldOptionList.tag("options") << SEMI,
).combine_dict(OneOfField)
oneof = seq(
    lexeme("oneof") >> oneofName.tag("name"),
    LBRACE
    >> (oneofField | emptyStatement).many().map(exclude_none).tag("fields")
    << RBRACE,
).combine_dict(OneOf)

# Map field
keyType = lexeme(from_enum(KeyType))
mapField = seq(
    lexeme("map") >> lexeme("<") >> keyType.tag("key_type"),
    lexeme(",") >> type_.tag("type"),
    lexeme(">") >> mapName.tag("name"),
    EQ >> fieldNumber.tag("number"),
    fieldOptionList.tag("options") << SEMI,
).combine_dict(Map)

# Reserved
range_ = seq(
    lexeme(intLit).tag("from_"),
    (lexeme("to") >> (intLit | lexeme("max"))).optional().tag("to"),
).combine_dict(Range)
ranges = range_.sep_by(lexeme(","), min=1)
# The spec for 'reserved' indicates 'fieldName' here, which is never a quoted string.
# But the example has a quoted string. We have changed it to 'strLit'
fieldNames = strLit.sep_by(lexeme(","), min=1)
reserved = seq(lexeme("reserved") >> (ranges | fieldNames) << SEMI).combine(Reserved)

# Enum definition
enumValueOption = seq(optionName.tag("name") << EQ, constant.tag("value")).combine_dict(
    Option
)
enumField = seq(
    ident.tag("name") << EQ,
    lexeme(intLit).tag("value"),
    (lexeme("[") >> enumValueOption.sep_by(lexeme(","), min=1) << lexeme("]"))
    .optional()
    .map(lambda o: [] if o is None else o)
    .tag("options")
    << SEMI,
).combine_dict(EnumField)
enumBody = (
    LBRACE >> (option | enumField | emptyStatement).many().map(exclude_none) << RBRACE
)
enum = seq(lexeme("enum") >> enumName.tag("name"), enumBody.tag("body")).combine_dict(
    Enum
)


# Message definition
@generate
def message():
    yield lexeme("message")
    name = yield messageName
    body = yield messageBody
    return Message(name=name, body=body)


messageBody = (
    LBRACE
    >> (
        field | enum | message | option | oneof | mapField | reserved | emptyStatement
    ).many()
    << RBRACE
)


# Service definition
rpc = seq(
    lexeme("rpc") >> rpcName.tag("name"),
    LPAREN >> (is_present("stream").tag("request_stream")),
    messageType.tag("request_message_type") << RPAREN,
    lexeme("returns") >> LPAREN >> (is_present("stream").tag("response_stream")),
    messageType.tag("response_message_type") << RPAREN,
    ((LBRACE >> (option | emptyStatement).many() << RBRACE) | SEMI.result([]))
    .optional()
    .map(exclude_none)
    .tag("options"),
).combine_dict(Rpc)

service = seq(
    lexeme("service") >> serviceName.tag("name"),
    LBRACE
    >> (option | rpc | emptyStatement).many().map(exclude_none).tag("body")
    << RBRACE,
).combine_dict(Service)


# Proto file
topLevelDef = message | enum | service
proto = seq(
    syntax.tag("syntax"),
    (import_ | package | option | topLevelDef | emptyStatement)
    .many()
    .map(exclude_none)
    .tag("statements"),
).combine_dict(Proto)


def parse_code(code: str) -> Proto:
    """Parse a proto code string

    Parameters
    ----------
    code : str. The proto code string.

    Returns
    -------
    result : Proto. The proto node structure.
    """
    return proto.parse(code)



# Copyright 2021 The KCL Authors. All rights reserved.

import typing
import copy
import os
import pathlib
from dataclasses import dataclass

import kclvm.kcl.ast as ast
import kclvm.kcl.types as types
import kclvm.kcl.error as kcl_error
import kclvm.unification as unification
import kclvm.compiler.astutil as astutil
import kclvm.compiler.parser as parser
from kclvm.tools.printer import PrintAST, Config
from kclvm.tools.format import kcl_fmt_file


KCL_FEATURE_GATE_OVERRIDE_AUTO_FIX_ENV = "KCL_FEATURE_GATE_OVERRIDE_AUTO_FIX"


@dataclass
class OverrideInfo:
    pkgpath: str = None
    filename: str = None
    module: ast.Module = None

    # ---------------
    # Static members
    # ---------------

    MODIFIED = []


class OverrideTransformer(ast.TreeTransformer):
    def __init__(
        self,
        target_id: str,
        field_path: str,
        override_key: ast.Identifier,
        override_value: ast.Literal,
        override_action: ast.OverrideAction = ast.OverrideAction.CREATE_OR_UPDATE,
    ):
        super().__init__()
        self.target_id: str = target_id
        self.field_path: str = field_path
        self.override_key: ast.Identifier = override_key
        self.override_value: ast.Literal = override_value
        self.override_target_count: int = 0
        self.has_override: bool = False
        self.override_action: ast.OverrideAction = override_action

    def walk_UnificationStmt(self, t: ast.UnificationStmt):
        """ast.AST: UnificationStmt

        Parameters
        ----------
        - target: Identifier
        - value: SchemaExpr
        """
        name = t.target.names[0]
        if name != self.target_id:
            return t
        self.override_target_count = 1
        self.has_override = True
        self.walk(t.value)
        return t

    def walk_AssignStmt(self, t: ast.AssignStmt):
        if not isinstance(t.value, ast.SchemaExpr):
            return t
        self.override_target_count = 0
        for target in t.targets:
            if not isinstance(target, ast.Identifier):
                continue
            assign_target = typing.cast(ast.Identifier, target)
            if len(assign_target.names) != 1:
                continue
            if assign_target.names[0] != self.target_id:
                continue
            self.override_target_count = self.override_target_count + 1
        if self.override_target_count == 0:
            return t
        self.has_override = True
        schema_expr_old: ast.SchemaExpr = copy.deepcopy(t.value)
        schema_expr_new: ast.SchemaExpr = t.value
        self.walk_SchemaExpr(schema_expr_new)

        if len(t.targets) == 1:
            return t

        # Fix multiple assign
        assign_stmt_list = []
        for target in t.targets:
            x: ast.AssignStmt = copy.deepcopy(t)
            x.targets = [target]
            if len(target.names) == 1 and target.names[0] == self.target_id:
                x.value = schema_expr_new
            else:
                x.value = schema_expr_old
            assign_stmt_list.append(x)
        return assign_stmt_list

    def walk_SchemaExpr(self, t: ast.SchemaExpr):
        if self.override_target_count <= 0:
            return t
        if not self._find_schema_config_and_replace(
            t, self.field_path, self.override_value
        ):
            # Not exist and append an override value when the action is CREATE_OR_UPDATE
            if self.override_action == ast.OverrideAction.CREATE_OR_UPDATE:
                t.config.items.append(
                    ast.ConfigEntry(
                        key=self.override_key,
                        value=self.override_value,
                        operation=ast.ConfigEntryOperation.OVERRIDE,
                    )
                )
        self.override_target_count = 0
        return t

    def walk_SchemaStmt(self, t: ast.SchemaStmt):
        """Do not override AssignStmt in SchemaStmt"""
        return t

    def _get_config_field_paths(
        self,
        config: typing.Union[ast.SchemaExpr, ast.ConfigExpr],
    ) -> typing.Tuple[typing.List[str], typing.List[str]]:
        def _get_key_value_paths(
            key, value
        ) -> typing.Tuple[typing.List[str], typing.List[str]]:
            _paths, _paths_with_id = [], []
            if isinstance(key, ast.Identifier):
                path = key.get_name()
            elif isinstance(key, ast.StringLit):
                path = key.value
            else:
                return _paths, _paths_with_id
            _paths.append(f"{path}")
            _paths_with_id.append(f"{path}")
            value_paths, value_paths_with_id = self._get_config_field_paths(value)
            if value_paths:
                _paths.extend([f"{path}.{value_path}" for value_path in value_paths])
                _paths_with_id.extend(
                    [f"{path}|{value_path}" for value_path in value_paths_with_id]
                )
            return _paths, _paths_with_id

        paths, paths_with_id = [], []
        if isinstance(config, ast.SchemaExpr):
            for item in config.config.items:
                _paths, _paths_with_id = _get_key_value_paths(item.key, item.value)
                paths.extend(_paths)
                paths_with_id.extend(_paths_with_id)
        elif isinstance(config, ast.ConfigExpr):
            for key, value in zip(config.keys, config.values):
                _paths, _paths_with_id = _get_key_value_paths(key, value)
                paths.extend(_paths)
                paths_with_id.extend(_paths_with_id)
        return paths, paths_with_id

    def _replace_with_id_path(
        self,
        schema_config: typing.Union[ast.SchemaExpr, ast.ConfigExpr],
        path_with_id: str,
        override_value: ast.Literal,
    ) -> typing.Optional[ast.SchemaExpr]:
        if not path_with_id or not schema_config:
            return None
        parts = path_with_id.split("|")
        config = schema_config

        def _get_path_from_key(key: ast.AST) -> str:
            path = ""
            if isinstance(key, ast.Identifier):
                path = key.get_name()
            elif isinstance(key, ast.StringLit):
                path = key.value
            return path

        for i, part in enumerate(parts):
            if isinstance(config, ast.SchemaExpr):
                delete_index_list = []
                config_ref = config
                for j, item in enumerate(config.config.items):
                    path = _get_path_from_key(item.key)
                    if path == part:
                        if self.override_action == ast.OverrideAction.CREATE_OR_UPDATE:
                            if i == len(parts) - 1:
                                override_value.set_ast_position(config)
                                item.value = override_value
                            config = item.value
                        elif self.override_action == ast.OverrideAction.DELETE:
                            delete_index_list.append(j)
                        continue
                config_ref.config.items = [
                    item
                    for j, item in enumerate(config_ref.config.items)
                    if j not in delete_index_list
                ]
            elif isinstance(config, ast.ConfigExpr):
                key_value_pairs = zip(config.keys, config.values)
                delete_index_list = []
                config_ref = config
                for j, key_value in enumerate(key_value_pairs):
                    key, value = key_value
                    path = _get_path_from_key(key)
                    if path == part:
                        if self.override_action == ast.OverrideAction.CREATE_OR_UPDATE:
                            if i == len(parts) - 1:
                                override_value.set_ast_position(config)
                                config.items[j].value = override_value
                            config = value
                        elif self.override_action == ast.OverrideAction.DELETE:
                            delete_index_list.append(j)
                        continue
                config_ref.items = [
                    item
                    for j, item in enumerate(config_ref.items)
                    if j not in delete_index_list
                ]
        if override_value:
            override_value.set_ast_position(config)
        return schema_config

    def _find_schema_config_and_replace(
        self, schema_config: ast.SchemaExpr, field_path: str, value: typing.Any
    ) -> bool:
        if not schema_config:
            raise Exception("override schema config can't be None")
        # Find field_path by nested identifier
        paths, paths_with_id = self._get_config_field_paths(schema_config)
        if field_path not in paths:
            return False
        path_with_id = paths_with_id[paths.index(field_path)]
        self._replace_with_id_path(schema_config, path_with_id, value)
        return True


def ApplyOverrides(
    prog: ast.Program,
    overrides: typing.List[ast.CmdOverrideSpec],
    import_paths: typing.List[str] = None,
):
    for override in overrides or []:
        pkgpath = override.pkgpath if override.pkgpath else prog.main
        if pkgpath in prog.pkgs:
            for mx in prog.pkgs[pkgpath]:
                if FixModuleOverride(mx, override):
                    OverrideInfo.MODIFIED.append(
                        OverrideInfo(
                            pkgpath=pkgpath,
                            filename=mx.GetFileName(root=prog.root),
                            module=mx,
                        )
                    )
                ModuleAddImportPaths(mx, import_paths)
    # Override type check and to auto fix
    if os.getenv(KCL_FEATURE_GATE_OVERRIDE_AUTO_FIX_ENV):
        types.ResolveProgram(prog, types.CheckConfig(config_attr_auto_fix=True))
    # Put AST module modified deepcopy
    OverrideInfo.MODIFIED = [copy.deepcopy(m) for m in OverrideInfo.MODIFIED]


def PrintOverridesAST(is_fix: bool = True):
    """Print override AST program"""
    if OverrideInfo.MODIFIED:
        for value in OverrideInfo.MODIFIED:
            with open(value.filename, "w") as f:
                f.flush()
                os.fsync(f.fileno())
                PrintAST(value.module, f, Config(is_fix=is_fix))
            kcl_fmt_file(pathlib.Path(value.filename))


def ModuleAddImportPaths(
    m: ast.Module, import_paths: typing.List[str], ignore_exist: bool = False
) -> ast.Module:
    """Add more import paths into the AST module."""
    if not import_paths:
        return m
    import_stmt_list = []
    exist_import_set = [
        f"{stmt.path} as {stmt.asname}" if stmt.asname else stmt.path
        for stmt in m.GetImportList()
    ]
    line = 1
    for path in import_paths or []:
        if not ignore_exist and path in exist_import_set:
            continue
        import_stmt = ast.ImportStmt(line, 1)
        import_stmt.path = path
        import_stmt.name = path.rsplit(".")[-1]
        import_stmt_list.append(import_stmt)
        line += 1
    m.body = import_stmt_list + m.body


def FixModuleOverride(m: ast.Module, override: ast.CmdOverrideSpec) -> bool:
    assert m
    assert override

    ss = override.field_path.split(".")
    if len(ss) <= 1:
        return False

    target_id: str = ss[0]
    field: str = ".".join(ss[1:])
    value: str = override.field_value

    key = ast.Identifier(names=[s for s in field.split(".")], ctx=ast.ExprContext.STORE)
    val = astutil.BuildNodeFromString(value)

    transformer = OverrideTransformer(
        target_id, field, key, val, override_action=override.action
    )
    transformer.walk(m)
    return transformer.has_override


def override_file(
    file: str, specs: typing.List[str], import_paths: typing.List[str] = None
) -> bool:
    """Override and rewrite a file with override spec

    Parameters
    ----------
    file: str
        The File that need to be overridden
    specs: List[str]
        List of specs that need to be overridden.
        Each spec string satisfies the form: <pkgpath>:<field_path>=<filed_value> or <pkgpath>:<field_path>-
        When the pkgpath is '__main__', it can be omitted.

    Return
    ------
    result: bool
        Whether override is successful
    """
    overrides = [spec_str_to_override(spec) for spec in specs or []]
    program = parser.LoadProgram(
        file,
        mode=parser.ParseMode.ParseComments,
        load_packages=bool(os.getenv(KCL_FEATURE_GATE_OVERRIDE_AUTO_FIX_ENV)),
    )
    # Config unification
    program.pkgs[ast.Program.MAIN_PKGPATH] = unification.MergeASTList(
        program.pkgs[ast.Program.MAIN_PKGPATH]
    )
    OverrideInfo.MODIFIED = []
    ApplyOverrides(program, overrides, import_paths)
    PrintOverridesAST(False)
    return True


def spec_str_to_override(spec: str) -> ast.CmdOverrideSpec:
    """Override spec string to override structure"""

    def report_exception():
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.IllegalArgumentError_TYPE,
            arg_msg=f"Invalid spec format '{spec}', expected <pkgpath>:<field_path>=<filed_value> or <pkgpath>:<field_path>-",
        )

    # Create or update the override value
    if "=" in spec:
        split_values = spec.split("=", 1)
        paths = split_values[0].split(":", 1)
        if len(split_values) < 2 or len(paths) > 2:
            report_exception()
        paths.append(split_values[1])
        paths = paths if len(paths) == 3 else ["", *paths]
        return ast.CmdOverrideSpec(
            pkgpath=paths[0],
            field_path=paths[1],
            field_value=paths[2],
            action=ast.OverrideAction.CREATE_OR_UPDATE,
        )
    # Delete the override value
    elif spec.endswith("-"):
        paths = spec[:-1].split(":", 1)
        if len(paths) > 2:
            report_exception()
        paths = paths if len(paths) == 2 else ["", *paths]
        return ast.CmdOverrideSpec(
            pkgpath=paths[0],
            field_path=paths[1],
            field_value="",
            action=ast.OverrideAction.DELETE,
        )

    report_exception()
